"""
This script retrieves bibliographic data from Wikidata based on a list of QIDs (from a CSV file)
and transforms it into RDF triples according to LRMoo/FRBRoo and CIDOC CRM (OWL/eCRM).

The output is serialized as Turtle and written to 'works.ttl'.

"""

import csv
import time
import logging
import requests
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry
from email.utils import parsedate_to_datetime
from datetime import datetime, timezone
from typing import Optional, Dict, Any, Iterable, List
from pathlib import Path
import argparse
from pyshacl import validate
from wiki2crm import resources

from rdflib import Graph, Literal, Namespace, URIRef
from rdflib.namespace import RDF, RDFS, OWL, XSD
from tqdm import tqdm

# Settings
SPARQL_ENDPOINT = "https://query.wikidata.org/sparql"
SESSION_USER_AGENT = "SapphoWorkIntegrationBot/1.0 (laura.untner@fu-berlin.de)"
HTTP_TIMEOUT = 90
MAX_RETRIES = 5

# Namespaces
CRM = Namespace("http://www.cidoc-crm.org/cidoc-crm/") # CIDOC CRM
ECRM = Namespace("http://erlangen-crm.org/current/") #eCRM – CIDOC CRM (OWL version)
ECRM_URI = URIRef("http://erlangen-crm.org/current/")
LRMOO = Namespace("http://iflastandards.info/ns/lrm/lrmoo/") # LRMoo
LRMOO_URI = URIRef("https://cidoc-crm.org/extensions/lrmoo/owl/1.0/LRMoo_v1.0.owl")
FRBROO = Namespace("http://iflastandards.info/ns/fr/frbr/frbroo/") # FRBRoo
EFRBROO = Namespace("http://erlangen-crm.org/efrbroo/") # eFRBRoo
PROV = Namespace("http://www.w3.org/ns/prov#") # PROV-O - Provenance Ontology
WD = "http://www.wikidata.org/entity/" # Base URI for Wikidata entities
SAPPHO_BASE_URI = Namespace("https://sappho-digital.com/")

# HTTP helpers (Retry-After aware)
def _parse_retry_after(header_val: str) -> Optional[float]:
    """
    Parse 'Retry-After' per RFC 7231. It may be either a delay in seconds
    or an HTTP-date. Returns seconds as float, or None if not parseable.
    """
    if not header_val:
        return None
    header_val = header_val.strip()
    if header_val.isdigit():
        return float(header_val)
    try:
        dt = parsedate_to_datetime(header_val)
        if dt.tzinfo is None:
            dt = dt.replace(tzinfo=timezone.utc)
        return max(0.0, (dt - datetime.now(timezone.utc)).total_seconds())
    except Exception:
        return None

def make_session() -> requests.Session:
    """
    Build a pooled Session. Manual retry handling allows full respect of Retry-After.
    """
    sess = requests.Session()
    sess.headers.update({"User-Agent": SESSION_USER_AGENT})
    retry = Retry(
        total=0,  # manual control below
        respect_retry_after_header=True,
        backoff_factor=0,
        status_forcelist=(429, 500, 502, 503, 504),
        allowed_methods=frozenset(["GET", "POST"]),
        raise_on_status=False,
    )
    adapter = HTTPAdapter(max_retries=retry, pool_connections=10, pool_maxsize=20)
    sess.mount("http://", adapter)
    sess.mount("https://", adapter)
    return sess

SESSION = make_session()

def http_request_with_retry(
    method: str,
    url: str,
    *,
    params: Optional[Dict[str, Any]] = None,
    data: Optional[Dict[str, Any]] = None,
    headers: Optional[Dict[str, str]] = None,
    ok_statuses: Iterable[int] = (200,),
    max_retries: int = MAX_RETRIES,
    timeout: int = HTTP_TIMEOUT,
) -> requests.Response:
    """
    Perform an HTTP request and retry on 429/5xx.
    Uses Retry-After if present; otherwise applies a gentle backoff.
    """
    tries = 0
    while True:
        tries += 1
        resp = SESSION.request(method, url, params=params, data=data, headers=headers, timeout=timeout)

        if resp.status_code in ok_statuses:
            return resp

        if resp.status_code == 429:
            retry_after = _parse_retry_after(resp.headers.get("Retry-After", ""))
            wait_s = retry_after if retry_after is not None else 5.0
            print(f"429 Too Many Requests – waiting {wait_s:.1f}s (try {tries}/{max_retries})")
            if tries >= max_retries:
                resp.raise_for_status()
            time.sleep(wait_s)
            continue

        if 500 <= resp.status_code < 600:
            retry_after = _parse_retry_after(resp.headers.get("Retry-After", "")) or min(10.0, 1.5 ** (tries - 1))
            print(f"{resp.status_code} Server error – waiting {retry_after:.1f}s (try {tries}/{max_retries})")
            if tries >= max_retries:
                resp.raise_for_status()
            time.sleep(retry_after)
            continue

        # non-retriable error
        resp.raise_for_status()

def query_wikidata_raw(query: str, user_agent: str, timeout: int = HTTP_TIMEOUT) -> dict:
    """
    Execute a SPARQL query against Wikidata using the retry-aware HTTP routine.
    'user_agent' allows mirroring the original per-call UA strings.
    """
    headers = {"Accept": "application/sparql-results+json", "User-Agent": user_agent}
    resp = http_request_with_retry(
        "GET",
        SPARQL_ENDPOINT,
        params={"query": query},
        headers=headers,
        timeout=timeout,
    )
    return resp.json()

# Graph construction
def create_graph() -> Graph:
    """
    Create the RDF Graph and reproduce the original ontology and alignments.
    """
    g = Graph()
    g.bind("crm", CRM)
    g.bind("ecrm", ECRM)
    g.bind("lrmoo", LRMOO)
    g.bind("frbroo", FRBROO)
    g.bind("efrbroo", EFRBROO)
    g.bind("prov", PROV)
    g.bind("owl", OWL)
    g.bind("rdfs", RDFS)
    g.bind("sappho", SAPPHO_BASE_URI)

    # Ontology
    ontology_uri = URIRef("https://sappho-digital.com/ontology/works")
    g.add((ontology_uri, RDF.type, OWL.Ontology))
    g.add((ontology_uri, OWL.imports, ECRM_URI))
    g.add((ontology_uri, OWL.imports, LRMOO_URI))

    # Ontology Alignments (ECRM - CRM, LRMoo - FRBRoo/eFRBRoo) and property inverses
    ecrm_to_crm = [
        # Classes
        ("E21_Person",),
        ("E35_Title",),
        ("E42_Identifier",),
        ("E52_Time-Span",),
        ("E53_Place",),
        ("E55_Type",),
        ("E62_String",),
        ("E73_Information_Object",),
        ("E82_Actor_Appellation",),
        ("E40_Legal_Body",),
    ]

    for cls in ecrm_to_crm:
        g.add((ECRM.term(cls[0]), OWL.equivalentClass, CRM.term(cls[0])))

    # Properties
    ecrm_properties = [
        ("P1_is_identified_by", "P1i_identifies"),
        ("P2_has_type", "P2i_is_type_of"),
        ("P4_has_time-span", "P4i_is_time-span_of"),
        ("P7_took_place_at", "P7i_witnessed"),
        ("P14_carried_out_by", "P14i_performed"),
        ("P102_has_title", "P102i_is_title_of"),
        ("P131_is_identified_by", "P131i_identifies"),
        ("P138_represents", "P138i_has_representation"),
        ("P190_has_symbolic_content", "P190i_is_content_of")
    ]

    for direct, inverse in ecrm_properties:
        g.add((ECRM.term(direct), OWL.inverseOf, ECRM.term(inverse)))
        g.add((ECRM.term(direct), OWL.equivalentProperty, CRM.term(direct)))
        g.add((ECRM.term(inverse), OWL.inverseOf, ECRM.term(direct)))
        g.add((ECRM.term(inverse), OWL.equivalentProperty, CRM.term(inverse)))

    lrmoo_to_frbroo = {
        "F1_Work": "F1_Work",
        "F2_Expression": "F2_Expression",
        "F3_Manifestation": "F3_Manifestation_Product_Type",
        "F5_Item": "F5_Item",
        "F27_Work_Creation": "F27_Work_Conception",
        "F28_Expression_Creation": "F28_Expression_Creation",
        "F30_Manifestation_Creation": "F30_Publication_Event",
        "F32_Item_Production_Event": "F32_Carrier_Production_Event"
    }
    for lr, fr in lrmoo_to_frbroo.items():
        g.add((LRMOO.term(lr), OWL.equivalentClass, FRBROO.term(fr)))
        g.add((LRMOO.term(lr), OWL.equivalentClass, EFRBROO.term(fr)))

    lrmoo_properties = [
        ("R3_is_realised_in", "R3i_realises", "R3_is_realised_in", "R3i_realises"),
        ("R4_embodies", "R4i_is_embodied_in", "R4i_comprises_carriers_of", "R4_carriers_provided_by"),
        ("R7_exemplifies", "R7i_is_exemplified_by", "R7_is_example_of", "R7i_has_example"),
        ("R16_created", "R16i_was_created_by", "R16_initiated", "R16i_was_initiated_by"),
        ("R17_created", "R17i_was_created_by", "R17_created", "R17i_was_created_by"),
        ("R19_created_a_realisation_of", "R19i_was_realised_through", "R19_created_a_realisation_of", "R19i_was_realised_through"),
        ("R24_created", "R24i_was_created_through", "R24_created", "R24i_was_created_through"),
        ("R27_materialized", "R27i_was_materialized_by", "R27_used_as_source_material", "R27i_was_used_by"),
        ("R28_produced", "R28i_was_produced_by", "R28_produced", "R28i_was_produced_by"),
    ]

    for lr_direct, lr_inverse, fr_direct, fr_inverse in lrmoo_properties:
        g.add((LRMOO.term(lr_direct), OWL.inverseOf, LRMOO.term(lr_inverse)))
        g.add((LRMOO.term(lr_inverse), OWL.inverseOf, LRMOO.term(lr_direct)))
        g.add((LRMOO.term(lr_direct), OWL.equivalentProperty, FRBROO.term(fr_direct)))
        g.add((LRMOO.term(lr_direct), OWL.equivalentProperty, EFRBROO.term(fr_direct)))
        g.add((LRMOO.term(lr_inverse), OWL.equivalentProperty, FRBROO.term(fr_inverse)))
        g.add((LRMOO.term(lr_inverse), OWL.equivalentProperty, EFRBROO.term(fr_inverse)))

    return g

# CSV & utilities
def load_qids(path: Path) -> List[str]:
    qids: List[str] = []
    with path.open() as f:
        for row in csv.reader(f):
            qid = row[0].strip()
            if qid.startswith("Q"):
                qids.append(qid)
    return qids

def extract_year(date_str: str) -> Optional[str]:
    return date_str[:4] if date_str else None

# SPARQL queries
def query_wikidata(qids: List[str], max_retries: int = MAX_RETRIES):
    values = " ".join(f"wd:{qid}" for qid in qids)
    query = f"""
    SELECT ?work ?workLabel ?title_de ?title_en ?genre ?genreLabel ?author ?authorLabel ?creation_date (MIN(?raw_pub_date) AS ?pub_date) ?pub_place ?pub_placeLabel ?publisher ?publisherLabel ?digitalCopy ?editor ?editorLabel ?publishedIn ?partOf WHERE {{
    VALUES ?work {{ {values} }}
    OPTIONAL {{ ?work wdt:P1476 ?title_de . FILTER(LANG(?title_de) = 'de') }}
    OPTIONAL {{ ?work wdt:P1476 ?title_en . FILTER(LANG(?title_en) = 'en') }}
    OPTIONAL {{ ?work wdt:P136 ?genre . }}
    OPTIONAL {{ ?work wdt:P50 ?author . }}
    OPTIONAL {{ ?work wdt:P577 ?raw_pub_date . }}
    OPTIONAL {{ ?work wdt:P291 ?pub_place . }}
    OPTIONAL {{ ?work wdt:P123 ?publisher . }}
    OPTIONAL {{ ?work wdt:P953 ?digitalCopy . }}
    OPTIONAL {{ ?work wdt:P98 ?editor . }}
    OPTIONAL {{ ?work wdt:P1433 ?publishedIn . }}
    OPTIONAL {{ ?work wdt:P361 ?partOf . }}
    OPTIONAL {{ {{ ?work wdt:P571 ?creation_date . }} UNION {{ ?work wdt:P2754 ?creation_date . }} }}
    SERVICE wikibase:label {{ bd:serviceParam wikibase:language "en,de" }}
    }}
    GROUP BY ?work ?workLabel ?title_de ?title_en ?genre ?genreLabel ?author ?authorLabel ?creation_date ?pub_place ?pub_placeLabel ?publisher ?publisherLabel ?digitalCopy ?editor ?editorLabel ?publishedIn ?partOf
    """
    ua = "SapphoWorkIntegrationBot/1.0 (laura.untner@fu-berlin.de)"
    tries = 0
    while True:
        tries += 1
        try:
            data = query_wikidata_raw(query, user_agent=ua, timeout=HTTP_TIMEOUT)
            return data["results"]["bindings"] if "results" in data else data.get("bindings", [])
        except Exception as e:
            wait = min(5.0 * tries, 20.0)
            print(f"[RETRY {tries}] SPARQL request failed: {e} – retrying in {wait:.1f}s...")
            if tries >= max_retries:
                print("[ERROR] Maximum retries reached. Skipping batch.")
                return []
            time.sleep(wait)

def fetch_label(qid: str):
    query = f"""
    SELECT ?label_de ?label_en WHERE {{
      wd:{qid} rdfs:label ?label_de . FILTER(LANG(?label_de) = "de")
      OPTIONAL {{ wd:{qid} rdfs:label ?label_en . FILTER(LANG(?label_en) = "en") }}
    }}
    """
    ua = "SapphoWorkIntegrationBot/1.0 (contact@example.com)"
    try:
        data = query_wikidata_raw(query, user_agent=ua, timeout=30)
        results = data["results"]["bindings"]
        if results:
            label_de = results[0].get("label_de", {}).get("value")
            label_en = results[0].get("label_en", {}).get("value")
            if label_de:
                return label_de, "de"
            elif label_en:
                return label_en, "en"
    except Exception as e:
        print(f"[WARN] Could not fetch label for {qid}: {e}")
    return "Untitled", "en"

# Label helpers
def label_for(title_de, title_en, work_label):
    if title_de:
        return title_de, "de"
    elif title_en:
        return title_en, "en"
    elif work_label:
        return work_label, "de"
    else:
        return "Untitled", "en"

def manifestation_label_for(r):
    if "publishedInLabel" in r:
        return r["publishedInLabel"]["value"], "de"
    elif "partOfLabel" in r:
        return r["partOfLabel"]["value"], "de"
    elif "publishedIn" in r:
        parent_qid = r["publishedIn"]["value"].split("/")[-1]
        return fetch_label(parent_qid)
    elif "partOf" in r:
        parent_qid = r["partOf"]["value"].split("/")[-1]
        return fetch_label(parent_qid)
    else:
        return None, None

# Main processing
def process(g: Graph, qids: List[str]) -> None:
    # Caches for deduplication
    genre_cache: Dict[str, URIRef] = {}
    date_cache: Dict[str, URIRef] = {}
    publisher_cache: Dict[str, URIRef] = {}
    place_cache: Dict[str, URIRef] = {}

    # Process in Batches of 20
    for i in tqdm(range(0, len(qids), 20)):
        batch = qids[i:i+20]
        results = query_wikidata(batch)
        for r in results:
            # Triple Creation
            qid = r["work"]["value"].split("/")[-1]
            work_uri = URIRef(f"{SAPPHO_BASE_URI}work/{qid}")
            expression_uri = URIRef(f"{SAPPHO_BASE_URI}expression/{qid}")
            label, lang = label_for(
                r.get("title_de", {}).get("value"),
                r.get("title_en", {}).get("value"),
                r.get("workLabel", {}).get("value")
            )
            title_uri = URIRef(f"{SAPPHO_BASE_URI}title/expression/{qid}")
            title_string_uri = URIRef(f"{SAPPHO_BASE_URI}title_string/expression/{qid}")

            g.add((work_uri, RDF.type, LRMOO.F1_Work))
            g.add((work_uri, RDFS.label, Literal(f"Work of {label}", lang="en")))
            g.add((work_uri, LRMOO.R3_is_realised_in, expression_uri))
            g.add((expression_uri, LRMOO.R3i_realises, work_uri))

            # Work Creation
            work_creation_uri = URIRef(f"{SAPPHO_BASE_URI}work_creation/{qid}")
            g.add((work_creation_uri, RDF.type, LRMOO.F27_Work_Creation))
            g.add((work_creation_uri, RDFS.label, Literal(f"Work creation of {label}", lang="en")))
            g.add((work_creation_uri, LRMOO.R16_created, work_uri))
            g.add((work_uri, LRMOO.R16i_was_created_by, work_creation_uri))
            g.add((work_creation_uri, PROV.wasDerivedFrom, URIRef(f"{WD}{qid}")))

            if "author" in r:
                author_qid = r["author"]["value"].split("/")[-1]
                author_uri = URIRef(f"{SAPPHO_BASE_URI}person/{author_qid}")
                g.add((work_creation_uri, ECRM.P14_carried_out_by, author_uri))
                g.add((author_uri, ECRM.P14i_performed, work_creation_uri))
                
                g.add((author_uri, RDF.type, ECRM.E21_Person))
                g.add((author_uri, RDFS.label, Literal(r.get("authorLabel", {}).get("value", "Unknown"))))
                g.add((author_uri, OWL.sameAs, URIRef(r["author"]["value"])))

            # Expression
            g.add((expression_uri, RDF.type, LRMOO.F2_Expression))
            g.add((expression_uri, RDFS.label, Literal(f"Expression of {label}", lang="en")))                
            identifier_uri = URIRef(f"{SAPPHO_BASE_URI}identifier/{qid}")
            g.add((expression_uri, ECRM.P1_is_identified_by, identifier_uri))
            g.add((identifier_uri, ECRM.P1i_identifies, expression_uri))
            g.add((identifier_uri, RDF.type, ECRM.E42_Identifier))
            g.add((identifier_uri, RDFS.label, Literal(qid)))
            g.add((identifier_uri, ECRM.P2_has_type, URIRef("https://sappho-digital.com/id_type/wikidata")))
            g.add((URIRef("https://sappho-digital.com/id_type/wikidata"), ECRM.P2i_is_type_of, identifier_uri))
            
            wikidata_id_type_uri = URIRef("https://sappho-digital.com/id_type/wikidata")
            g.add((wikidata_id_type_uri, RDF.type, ECRM.E55_Type))
            g.add((wikidata_id_type_uri, RDFS.label, Literal("Wikidata ID", lang="en")))
            g.add((wikidata_id_type_uri, OWL.sameAs, URIRef("http://www.wikidata.org/wiki/Q43649390")))
            g.add((expression_uri, ECRM.P102_has_title, title_uri))
            g.add((title_uri, ECRM.P102i_is_title_of, expression_uri))
            g.add((title_uri, RDF.type, ECRM.E35_Title))
            g.add((title_uri, ECRM.P190_has_symbolic_content, title_string_uri))
            g.add((title_string_uri, ECRM.P190i_is_content_of, title_uri))
            g.add((title_string_uri, RDF.type, ECRM.E62_String))
            g.add((title_string_uri, RDFS.label, Literal(label, lang=lang)))

            if "genre" in r:
                genre_qid = r["genre"]["value"].split("/")[-1]
                if genre_qid not in genre_cache:
                    genre_uri = URIRef(f"{SAPPHO_BASE_URI}genre/{genre_qid}")
                    genre_cache[genre_qid] = genre_uri
                    g.add((genre_uri, RDF.type, ECRM.E55_Type))
                    g.add((genre_uri, RDFS.label, Literal(r.get("genreLabel", {}).get("value", "Unknown"), lang="en")))
                    g.add((genre_uri, OWL.sameAs, URIRef(r["genre"]["value"])))
                    g.add((genre_uri, ECRM.P2_has_type, URIRef(f"{SAPPHO_BASE_URI}genre_type/wikidata")))
                    g.add((URIRef(f"{SAPPHO_BASE_URI}genre_type/wikidata"), ECRM.P2i_is_type_of, genre_uri))
                    g.add((URIRef(f"{SAPPHO_BASE_URI}genre_type/wikidata"), RDF.type, ECRM.E55_Type))
                    g.add((URIRef(f"{SAPPHO_BASE_URI}genre_type/wikidata"), RDFS.label, Literal("Wikidata Genre", lang="en")))
                g.add((expression_uri, ECRM.P2_has_type, genre_cache[genre_qid]))
                g.add((genre_cache[genre_qid], ECRM.P2i_is_type_of, expression_uri))

            g.add((expression_uri, OWL.sameAs, URIRef(f"{WD}{qid}")))
            g.add((expression_uri, PROV.wasDerivedFrom, URIRef(f"{WD}{qid}")))

            # Expression Creation
            expression_creation_uri = URIRef(f"{SAPPHO_BASE_URI}expression_creation/{qid}")
            g.add((expression_creation_uri, RDF.type, LRMOO.F28_Expression_Creation))
            g.add((expression_creation_uri, RDFS.label, Literal(f"Expression creation of {label}", lang="en")))
            g.add((expression_creation_uri, LRMOO.R17_created, expression_uri))
            g.add((expression_uri, LRMOO.R17i_was_created_by, expression_creation_uri))
            g.add((expression_creation_uri, LRMOO.R19_created_a_realisation_of, work_uri))
            g.add((work_uri, LRMOO.R19i_was_realised_through, expression_creation_uri))
            g.add((expression_creation_uri, PROV.wasDerivedFrom, URIRef(f"{WD}{qid}")))

            if "author" in r:
                g.add((expression_creation_uri, ECRM.P14_carried_out_by, author_uri))

            if "creation_date" in r:
                year = extract_year(r["creation_date"]["value"])
                if year:
                    if year not in date_cache:
                        date_uri = URIRef(f"{SAPPHO_BASE_URI}timespan/{year}")
                        date_cache[year] = date_uri
                        g.add((date_uri, RDF.type, ECRM["E52_Time-Span"]))
                        g.add((date_uri, RDFS.label, Literal(year, datatype=XSD.gYear)))
                    g.add((expression_creation_uri, ECRM["P4_has_time-span"], date_cache[year]))
                    g.add((date_cache[year], ECRM["P4i_is_time-span_of"], expression_creation_uri))

            # Manifestation
            manifestation_uri = URIRef(f"{SAPPHO_BASE_URI}manifestation/{qid}")
            g.add((manifestation_uri, RDF.type, LRMOO.F3_Manifestation))
            g.add((manifestation_uri, RDFS.label, Literal(f"Manifestation of {label}", lang="en")))
            g.add((manifestation_uri, LRMOO.R4_embodies, expression_uri))
            g.add((expression_uri, LRMOO.R4i_is_embodied_in, manifestation_uri))
            
            manifestation_title_uri = URIRef(f"{SAPPHO_BASE_URI}title/manifestation/{qid}")
            manifestation_title_string_uri = URIRef(f"{SAPPHO_BASE_URI}title_string/manifestation/{qid}")
            
            manifestation_label, manifestation_lang = manifestation_label_for(r)
            if manifestation_label is None:
                manifestation_label = label
                manifestation_lang = lang
            
            if "publishedIn" in r:
                parent_qid = r["publishedIn"]["value"].split("/")[-1]
                manifestation_label, manifestation_lang = fetch_label(parent_qid)
            elif "partOf" in r:
                parent_qid = r["partOf"]["value"].split("/")[-1]
                manifestation_label, manifestation_lang = fetch_label(parent_qid)

            g.add((manifestation_uri, ECRM.P102_has_title, manifestation_title_uri))
            g.add((manifestation_title_uri, ECRM.P102i_is_title_of, manifestation_uri))
            g.add((manifestation_title_uri, RDF.type, ECRM.E35_Title))
            g.add((manifestation_title_uri, ECRM.P190_has_symbolic_content, manifestation_title_string_uri))
            g.add((manifestation_title_string_uri, ECRM.P190i_is_content_of, manifestation_title_uri))
            g.add((manifestation_title_string_uri, RDF.type, ECRM.E62_String))
            g.add((manifestation_title_string_uri, RDFS.label, Literal(manifestation_label, lang=manifestation_lang)))

            # Manifestation Creation
            manifestation_creation_uri = URIRef(f"{SAPPHO_BASE_URI}manifestation_creation/{qid}")
            g.add((manifestation_creation_uri, RDF.type, LRMOO.F30_Manifestation_Creation))
            g.add((manifestation_creation_uri, RDFS.label, Literal(f"Manifestation creation of {label}", lang="en")))
            g.add((manifestation_creation_uri, LRMOO.R24_created, manifestation_uri))
            g.add((manifestation_uri, LRMOO.R24i_was_created_through, manifestation_creation_uri))
            g.add((manifestation_creation_uri, PROV.wasDerivedFrom, URIRef(f"{WD}{qid}")))
            if "author" in r:
                g.add((manifestation_creation_uri, ECRM.P14_carried_out_by, author_uri))
                g.add((author_uri, ECRM.P14i_performed, manifestation_creation_uri))

            if "publisher" in r:
                publisher_qid = r["publisher"]["value"].split("/")[-1]
                if publisher_qid not in publisher_cache:
                    publisher_uri = URIRef(f"{SAPPHO_BASE_URI}publisher/{publisher_qid}")
                    publisher_cache[publisher_qid] = publisher_uri
                    g.add((publisher_uri, RDF.type, ECRM.E40_Legal_Body))
                    g.add((publisher_uri, RDFS.label, Literal(r.get("publisherLabel", {}).get("value", "Unknown"), lang="en")))
                    g.add((publisher_uri, OWL.sameAs, URIRef(r["publisher"]["value"])))
                g.add((manifestation_creation_uri, ECRM.P14_carried_out_by, publisher_cache[publisher_qid]))
                g.add((publisher_cache[publisher_qid], ECRM.P14i_performed, manifestation_creation_uri))

            if "pub_date" in r:
                pub_year = extract_year(r["pub_date"]["value"])
                if pub_year:
                    if pub_year not in date_cache:
                        pub_date_uri = URIRef(f"{SAPPHO_BASE_URI}timespan/{pub_year}")
                        date_cache[pub_year] = pub_date_uri
                        g.add((pub_date_uri, RDF.type, ECRM["E52_Time-Span"]))
                        g.add((pub_date_uri, RDFS.label, Literal(pub_year, datatype=XSD.gYear)))
                    g.add((manifestation_creation_uri, ECRM["P4_has_time-span"], date_cache[pub_year]))
                    g.add((date_cache[pub_year], ECRM["P4i_is_time-span_of"], manifestation_creation_uri))

            if "pub_place" in r:
                place_qid = r["pub_place"]["value"].split("/")[-1]
                if place_qid not in place_cache:
                    place_uri = URIRef(f"{SAPPHO_BASE_URI}place/{place_qid}")
                    place_cache[place_qid] = place_uri
                    g.add((place_uri, RDF.type, ECRM.E53_Place))
                    g.add((place_uri, RDFS.label, Literal(r.get("pub_placeLabel", {}).get("value", "Unknown"), lang="en")))
                    g.add((place_uri, OWL.sameAs, URIRef(r["pub_place"]["value"])))
                g.add((manifestation_creation_uri, ECRM.P7_took_place_at, place_cache[place_qid]))
                g.add((place_cache[place_qid], ECRM.P7i_witnessed, manifestation_creation_uri))

            if "editor" in r:
                editor_qid = r["editor"]["value"].split("/")[-1]
                editor_uri = URIRef(f"{SAPPHO_BASE_URI}editor/{editor_qid}")
                g.add((editor_uri, RDF.type, ECRM.E21_Person))
                g.add((editor_uri, RDFS.label, Literal(r.get("editorLabel", {}).get("value", "Unknown"))))
                g.add((editor_uri, OWL.sameAs, URIRef(r["editor"]["value"])))

                app_uri = URIRef(f"{SAPPHO_BASE_URI}appellation/{editor_qid}")
                g.add((editor_uri, ECRM.P131_is_identified_by, app_uri))
                g.add((app_uri, ECRM.P131i_identifies, editor_uri))
                g.add((app_uri, RDF.type, ECRM.E82_Actor_Appellation))
                g.add((app_uri, RDFS.label, Literal(r.get("editorLabel", {}).get("value", "Unknown"))))
                g.add((app_uri, PROV.wasDerivedFrom, URIRef(r["editor"]["value"])))

                id_uri = URIRef(f"{SAPPHO_BASE_URI}identifier/{editor_qid}")
                g.add((editor_uri, ECRM.P1_is_identified_by, id_uri))
                g.add((id_uri, ECRM.P1i_identifies, editor_uri))
                g.add((id_uri, RDF.type, ECRM.E42_Identifier))
                g.add((id_uri, RDFS.label, Literal(editor_qid)))
                g.add((id_uri, ECRM.P2_has_type, URIRef("https://sappho-digital.com/id_type/wikidata")))
                g.add((URIRef("https://sappho-digital.com/id_type/wikidata"), ECRM.P2i_is_type_of, id_uri))
                g.add((manifestation_creation_uri, ECRM.P14_carried_out_by, editor_uri))
                g.add((editor_uri, ECRM.P14i_performed, manifestation_creation_uri))

            item_production_uri = URIRef(f"{SAPPHO_BASE_URI}item_production/{qid}")
            item_uri = URIRef(f"{SAPPHO_BASE_URI}item/{qid}")

            g.add((item_production_uri, RDF.type, LRMOO.F32_Item_Production_Event))
            g.add((item_production_uri, RDFS.label, Literal(f"Item production event of {label}", lang="en")))
            g.add((item_production_uri, LRMOO.R27_materialized, manifestation_uri))
            g.add((manifestation_uri, LRMOO.R27i_was_materialized_by, item_production_uri))
            g.add((item_production_uri, LRMOO.R28_produced, item_uri))
            g.add((item_uri, LRMOO.R28i_was_produced_by, item_production_uri))

            g.add((item_uri, RDF.type, LRMOO.F5_Item))
            g.add((item_uri, RDFS.label, Literal(f"Item of {label}", lang="en")))
            g.add((item_uri, LRMOO.R7_exemplifies, manifestation_uri))
            g.add((manifestation_uri, LRMOO.R7i_is_exemplified_by, item_uri))

            if "digitalCopy" in r:
                digital_uri = URIRef(f"{SAPPHO_BASE_URI}digital/{qid}")
                g.add((digital_uri, RDF.type, ECRM.E73_Information_Object))
                g.add((digital_uri, RDFS.label, Literal(f"Digital copy of {label}", lang="en")))
                g.add((digital_uri, ECRM.P138_represents, expression_uri))
                g.add((expression_uri, ECRM.P138i_has_representation, digital_uri))
                g.add((digital_uri, RDFS.seeAlso, URIRef(r["digitalCopy"]["value"])))

def parse_args(argv=None) -> argparse.Namespace:
    p = argparse.ArgumentParser(
        description="Retrieve bibliographic data from Wikidata and emit LRMoo/FRBRoo + CIDOC CRM TTL"
    )
    p.add_argument("--input",  type=Path, help="CSV with QIDs (e.g. examples/inputs/work-qids.csv)")
    p.add_argument("--output", type=Path, help="Output Turtle (e.g. examples/outputs/works.ttl)")
    p.add_argument(
        "--shapes",
        type=Path,
        default=resources.shapes_path("work-shapes.ttl"),
        help="Path to SHACL shapes (default: package-installed work-shapes.ttl)",
    )
    p.add_argument("--log-level", default="INFO", help="Logging level (default: INFO)")
    return p.parse_args(argv)

def main(argv=None) -> int:
    args = parse_args(argv)
    logging.basicConfig(level=getattr(logging, args.log_level.upper(), logging.INFO),
                        format="%(levelname)s:%(name)s:%(message)s")

    if args.input is None:
        repo_input = Path("examples/inputs/work-qids.csv")
        if repo_input.exists():
            args.input = repo_input
        else:
            raise SystemExit("--input is required (no examples/inputs/work-qids.csv found)")

    if args.output is None:
        repo_outdir = Path("examples/outputs")
        args.output = (repo_outdir / "works.ttl") if repo_outdir.exists() else Path("works.ttl")

    args.output.parent.mkdir(parents=True, exist_ok=True)

    # Create graph
    g = create_graph()

    # Load QIDs
    all_qids = load_qids(args.input)

    # Process and enrich graph
    process(g, all_qids)

    # Serialize to Turtle
    g.serialize(destination=str(args.output), format="turtle")
    print(f"✅ RDF graph written to {args.output}")

    # Validate the output graph using pySHACL
    shapes_graph = Graph().parse(str(args.shapes), format="turtle")

    conforms, report_graph, report_text = validate(
        g,
        shacl_graph=shapes_graph,
        inference="rdfs",
        meta_shacl=True,
        advanced=True,
        abort_on_first=False,
        debug=False,
    )

    print("\n🔎 SHACL Validation Report:")
    print(report_text)

    if not conforms:
        print("❌ Validation failed.")
    else:
        print("✅ Data conforms to SHACL shapes.")

    return 0

if __name__ == "__main__":
    raise SystemExit(main())


"""
This script merges ../authors/authors.ttl, ../works/works.ttl and ../relations/relations.ttl 
without duplications.

The output is serialized as Turtle and written to 'all.ttl'.

"""

import argparse
from collections import defaultdict
from pathlib import Path
from rdflib import Graph, Namespace, RDF, RDFS, OWL, URIRef

# Namespaces
sappho   = Namespace("https://sappho-digital.com/")
ecrm     = Namespace("http://erlangen-crm.org/current/")
ecrm_uri = URIRef("http://erlangen-crm.org/current/")
crm      = Namespace("http://www.cidoc-crm.org/cidoc-crm/")
lrmoo    = Namespace("http://iflastandards.info/ns/lrm/lrmoo/")
lrmoo_uri = URIRef("https://cidoc-crm.org/extensions/lrmoo/owl/1.0/LRMoo_v1.0.owl")
frbroo   = Namespace("http://iflastandards.info/ns/fr/frbr/frbroo/")
efrbroo  = Namespace("http://erlangen-crm.org/efrbroo/")
intro    = Namespace("https://w3id.org/lso/intro/currentbeta#")
intro_uri = URIRef("https://w3id.org/lso/intro/currentbeta#")
prov     = Namespace("http://www.w3.org/ns/prov#")

def load_graphs(authors_path: Path, works_path: Path, relations_path: Path) -> tuple[Graph, Graph, Graph]:
    """Load authors, works, and relations graphs from Turtle files."""
    g_authors = Graph().parse(str(authors_path), format="turtle")
    g_works = Graph().parse(str(works_path), format="turtle")
    g_relations = Graph().parse(str(relations_path), format="turtle")
    return g_authors, g_works, g_relations

def merge_graphs(graphs: list[Graph]) -> Graph:
    """Union all graphs into a new Graph (rdflib '+' semantics)."""
    g_all = Graph()
    for g in graphs:
        g_all += g
    return g_all

def cleanup_duplicate_labels(g_all: Graph) -> Graph:
    """
    Remove duplicate rdfs:label per subject.
    Keep exactly one label per subject: prefer a label with a language tag; 
    if multiple such labels exist, keep the first encountered; otherwise keep the first label.
    """
    label_map: dict[URIRef, list] = defaultdict(list)
    for s, p, o in g_all.triples((None, RDFS.label, None)):
        label_map[s].append(o)

    multi_label_subjects = {s for s, labels in label_map.items() if len(labels) > 1}

    cleaned_graph = Graph()
    for s, p, o in g_all:
        if p == RDFS.label and s in multi_label_subjects:
            labels = label_map[s]
            with_lang = [lbl for lbl in labels if getattr(lbl, "language", None)]
            keep = with_lang[0] if with_lang else labels[0]
            if o == keep:
                cleaned_graph.add((s, p, o))
        else:
            cleaned_graph.add((s, p, o))
    return cleaned_graph

def cleanup_ontology(g: Graph) -> Graph:
    """
    Remove any existing owl:Ontology nodes, then add the merged ontology node and imports.
    """
    for s, p, o in list(g.triples((None, RDF.type, OWL.Ontology))):
        g.remove((s, None, None))

    ontology_uri = URIRef("https://sappho-digital.com/ontology/all")
    g.add((ontology_uri, RDF.type, OWL.Ontology))
    g.add((ontology_uri, OWL.imports, ecrm_uri))
    g.add((ontology_uri, OWL.imports, lrmoo_uri))
    g.add((ontology_uri, OWL.imports, intro_uri))
    return g

def bind_namespaces(g: Graph) -> None:
    """Bind all project namespaces."""
    g.bind("sappho", sappho)
    g.bind("ecrm", ecrm)
    g.bind("crm", crm)
    g.bind("lrmoo", lrmoo)
    g.bind("frbroo", frbroo)
    g.bind("efrbroo", efrbroo)
    g.bind("intro", intro)
    g.bind("prov", prov)

def parse_args(argv=None) -> argparse.Namespace:
    p = argparse.ArgumentParser(
        description="Merge authors.ttl, works.ttl, relations.ttl into one graph (de-duplicated labels)"
    )
    p.add_argument("--authors",   type=Path, help="Path to authors.ttl (e.g. examples/outputs/authors.ttl)")
    p.add_argument("--works",     type=Path, help="Path to works.ttl (e.g. examples/outputs/works.ttl)")
    p.add_argument("--relations", type=Path, help="Path to relations.ttl (e.g. examples/outputs/relations.ttl)")
    p.add_argument("--output",    type=Path, help="Output TTL (e.g. examples/outputs/all.ttl)")
    p.add_argument("--log-level", default="INFO", help="Logging level (default: INFO)")
    return p.parse_args(argv)

def main(argv=None) -> int:
    args = parse_args(argv)

    import logging
    logging.basicConfig(level=getattr(logging, args.log_level.upper(), logging.INFO),
                        format="%(levelname)s:%(name)s:%(message)s")

    if args.authors is None:
        cand = Path("examples/outputs/authors.ttl")
        if cand.exists(): args.authors = cand
        else: raise SystemExit("--authors is required (no examples/outputs/authors.ttl found)")

    if args.works is None:
        cand = Path("examples/outputs/works.ttl")
        if cand.exists(): args.works = cand
        else: raise SystemExit("--works is required (no examples/outputs/works.ttl found)")

    if args.relations is None:
        cand = Path("examples/outputs/relations.ttl")
        if cand.exists(): args.relations = cand
        else: raise SystemExit("--relations is required (no examples/outputs/relations.ttl found)")

    if args.output is None:
        outdir = Path("examples/outputs")
        args.output = (outdir / "all.ttl") if outdir.exists() else Path("all.ttl")

    args.output.parent.mkdir(parents=True, exist_ok=True)

    for name, p in (("authors", args.authors), ("works", args.works), ("relations", args.relations)):
        if not p.exists():
            raise SystemExit(f"--{name} not found: {p}")

    # Merge
    g_authors, g_works, g_relations = load_graphs(args.authors, args.works, args.relations)
    g_all = merge_graphs([g_authors, g_works, g_relations])
    cleaned = cleanup_duplicate_labels(g_all)
    cleaned = cleanup_ontology(cleaned)
    bind_namespaces(cleaned)

    cleaned.serialize(destination=str(args.output), format="turtle")
    print(f"✅ merged TTL written to {args.output}")
    return 0

if __name__ == "__main__":
    raise SystemExit(main())


# src/wiki2crm/cli.py
import sys
import argparse
from importlib import import_module
from . import __version__

MODMAP = {
    "authors":   "wiki2crm.authors",
    "works":     "wiki2crm.works",
    "relations": "wiki2crm.relations",
    "merge":     "wiki2crm.merge",
    "map-align": "wiki2crm.map_and_align",
}

HELP_BY_CMD = {
    "authors":   "Build authors.ttl from a CSV of QIDs",
    "works":     "Build works.ttl from a CSV of QIDs",
    "relations": "Build relations.ttl from a CSV of QIDs",
    "merge":     "Merge authors/works/relations TTL into one",
    "map-align": "Add extra IDs + ontology alignments",
}

EXAMPLES = """Examples:
  wiki2crm authors --input /path/authors.csv --output /path/authors.ttl
  wiki2crm merge --authors a.ttl --works w.ttl --relations r.ttl --output all.ttl
  wiki2crm map-align --input all.ttl --output all_mapped-and-aligned.ttl
"""

def main(argv=None):
    argv = sys.argv[1:] if argv is None else list(argv)

    parser = argparse.ArgumentParser(
        prog="wiki2crm",
        description="Wikidata → CIDOC CRM",
        epilog=EXAMPLES,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument("--version", action="version", version=f"%(prog)s {__version__}")
    parser.add_argument("--hello", action="store_true", help="Print a short greeting and exit")

    sub = parser.add_subparsers(dest="cmd")
    for cmd, help_text in HELP_BY_CMD.items():
        sub.add_parser(cmd, help=help_text)

    args, rest = parser.parse_known_args(argv)

    if args.hello:
        print("👋 hello from wiki2crm")
        return 0

    if not args.cmd:
        parser.print_help()
        return 0

    modname = MODMAP.get(args.cmd)
    if not modname:
        parser.error(f"Unknown command: {args.cmd}")

    try:
        mod = import_module(modname)
    except Exception as e:
        parser.error(f"Could not import subcommand module '{modname}': {e}")

    if not hasattr(mod, "main"):
        parser.error(f"Subcommand module '{modname}' has no function main(argv=None)")

    return mod.main(argv=rest)

if __name__ == "__main__":
    raise SystemExit(main())



from importlib.metadata import version, PackageNotFoundError

try:
    from importlib.metadata import version, PackageNotFoundError
except Exception:
    def version(_): return "0+unknown"
    class PackageNotFoundError(Exception): pass

try:
    __version__ = version("wiki2crm")
except PackageNotFoundError:
    __version__ = "0+unknown"


"""
This script adds more ontology alignments to ../authors/authors.ttl, ../works/works.ttl, ../relations/relations.ttl or ../merge/all.ttl 
and adds new identifiers from DBpedia, GeoNames etc.

"""

from rdflib import Graph, Namespace, URIRef, BNode, Literal
from rdflib.namespace import RDF, RDFS, OWL
from rdflib.collection import Collection
import sys
import re
import time
from typing import Dict, Any, Iterable, Optional
import requests
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry
from email.utils import parsedate_to_datetime
from datetime import datetime, timezone
import argparse
from pathlib import Path

# Namespaces 
SKOS = Namespace("http://www.w3.org/2004/02/skos/core#")
CRM = Namespace("http://www.cidoc-crm.org/cidoc-crm/")
ECRM = Namespace("http://erlangen-crm.org/current/")
FRBROO = Namespace("http://iflastandards.info/ns/fr/frbr/frbroo/")
EFRBROO = Namespace("http://erlangen-crm.org/efrbroo/")
LRMOO = Namespace("http://iflastandards.info/ns/lrm/lrmoo/")
INTRO = Namespace("https://w3id.org/lso/intro/currentbeta#")
PROV = Namespace("http://www.w3.org/ns/prov#")
SAPPHO   = Namespace("https://sappho-digital.com/")
SAPPHO_PROP   = Namespace("https://sappho-digital.com/property/")

BIBO = Namespace("http://purl.org/ontology/bibo/")
CITO = Namespace("http://purl.org/spar/cito/")
DC = Namespace("http://purl.org/dc/terms/")
DOCO = Namespace("http://purl.org/spar/doco/")
FABIO = Namespace("http://purl.org/spar/fabio/")
FOAF = Namespace("http://xmlns.com/foaf/0.1/")
GOLEM = Namespace("https://ontology.golemlab.eu/")
DRACOR = Namespace("http://dracor.org/ontology#")
INTERTEXT_AB = Namespace("https://intertextuality.org/abstract#")
INTERTEXT_TX = Namespace("https://intertextuality.org/extensions/text#")
INTERTEXT_AF = Namespace("https://intertextuality.org/extensions/artifacts#")
INTERTEXT_MT = Namespace("https://intertextuality.org/extensions/motifs#")
MIMOTEXT = Namespace("http://data.mimotext.uni-trier.de/entity/")
ONTOPOETRY_CORE = Namespace("http://postdata.linhd.uned.es/ontology/postdata-core#")
ONTOPOETRY_ANALYSIS = Namespace("http://postdata.linhd.uned.es/ontology/postdata-poeticAnalysis#")
SCHEMA = Namespace("https://schema.org/")
URW = Namespace("https://purl.archive.org/urwriters#")
URB = Namespace("https://purl.archive.org/urbooks#")

# HTTP helpers
SPARQL_URL = "https://query.wikidata.org/sparql"
USER_AGENT = "SapphoMapAndAlignBot/1.0 (laura.untner@fu-berlin.de)"
HTTP_TIMEOUT = 120
MAX_RETRIES = 5

def _parse_retry_after(header_val: str) -> Optional[float]:
    if not header_val:
        return None
    header_val = header_val.strip()
    if header_val.isdigit():
        return float(header_val)
    try:
        dt = parsedate_to_datetime(header_val)
        if dt.tzinfo is None:
            dt = dt.replace(tzinfo=timezone.utc)
        return max(0.0, (dt - datetime.now(timezone.utc)).total_seconds())
    except Exception:
        return None

def _make_session() -> requests.Session:
    sess = requests.Session()
    sess.headers.update({"Accept": "application/sparql-results+json", "User-Agent": USER_AGENT})
    retry = Retry(total=0, respect_retry_after_header=True, backoff_factor=0,
                  status_forcelist=(429, 500, 502, 503, 504),
                  allowed_methods=frozenset(["GET"]), raise_on_status=False)
    adapter = HTTPAdapter(max_retries=retry, pool_connections=10, pool_maxsize=20)
    sess.mount("https://", adapter); sess.mount("http://", adapter)
    return sess

_SESSION = _make_session()

def _sparql_query(query: str, *, max_retries: int = MAX_RETRIES) -> Dict[str, Any]:
    tries = 0
    while True:
        tries += 1
        resp = _SESSION.get(SPARQL_URL, params={"query": query}, timeout=HTTP_TIMEOUT)
        if resp.status_code == 200:
            return resp.json()
        if resp.status_code == 429:
            ra = _parse_retry_after(resp.headers.get("Retry-After", "")) or 5.0
            print(f"429 Too Many Requests – waiting {ra:.1f}s (try {tries}/{max_retries})")
            if tries >= max_retries:
                resp.raise_for_status()
            time.sleep(ra); continue
        if 500 <= resp.status_code < 600:
            ra = _parse_retry_after(resp.headers.get("Retry-After", "")) or min(10.0, 1.5 ** (tries - 1))
            print(f"{resp.status_code} Server error – waiting {ra:.1f}s (try {tries}/{max_retries})")
            if tries >= max_retries:
                resp.raise_for_status()
            time.sleep(ra); continue
        resp.raise_for_status()

# Mapping

def normalize_uri(raw, prefix_map):
    if raw.startswith("http://") or raw.startswith("https://"):
        return raw
    if ":" in raw:
        prefix, local = raw.split(":", 1)
        base = prefix_map.get(prefix)
        if base:
            return base + local
    return raw

def query_wikidata_batch(qids): 
    prefix_map = {
        "schema":  "https://schema.org/",
        "dbpedia": "https://dbpedia.org/"
    }

    values = " ".join(f"wd:{qid}" for qid in qids)
    query = f"""
    PREFIX wd:  <http://www.wikidata.org/entity/>
    PREFIX wdt: <http://www.wikidata.org/prop/direct/>

    SELECT ?item ?schemaOrg ?dbpedia ?gnd ?viaf ?geonames ?grWork WHERE {{
      VALUES ?item {{ {values} }}
      OPTIONAL {{
        ?item wdt:P2888 | wdt:P1709 ?link1 .
        FILTER(STRSTARTS(STR(?link1),"https://schema.org/"))
        BIND(STR(?link1) AS ?schemaOrg)
      }}
      OPTIONAL {{
        ?item wdt:P2888 | wdt:P1709 ?link2 .
        FILTER(STRSTARTS(STR(?link2),"https://dbpedia.org/"))
        BIND(STR(?link2) AS ?dbpedia)
      }}
      OPTIONAL {{ ?item wdt:P227  ?gnd      }}
      OPTIONAL {{ ?item wdt:P214  ?viaf     }}
      OPTIONAL {{ ?item wdt:P1566 ?geonames }}
      OPTIONAL {{ ?item wdt:P8383 ?grWork   }}
    }}
    """
    results = _sparql_query(query)

    batch_ids = {
        qid: { 'schema':[], 'dbpedia':[], 'gnd':[], 'viaf':[], 'geonames':[], 'goodreads':[] }
        for qid in qids
    }

    for row in results['results']['bindings']:
        uri_item = row['item']['value']
        qid = uri_item.rsplit("/", 1)[1]

        if 'schemaOrg' in row:
            raw = row['schemaOrg']['value']
            batch_ids[qid]['schema'].append(normalize_uri(raw, prefix_map))

        if 'dbpedia' in row:
            raw = row['dbpedia']['value']
            batch_ids[qid]['dbpedia'].append(normalize_uri(raw, prefix_map))

        if 'gnd' in row:
            batch_ids[qid]['gnd'].append(f"http://d-nb.info/gnd/{row['gnd']['value']}")
        
        if 'viaf' in row:
            batch_ids[qid]['viaf'].append(f"http://viaf.org/viaf/{row['viaf']['value']}")
        
        if 'geonames' in row:
            batch_ids[qid]['geonames'].append(f"http://sws.geonames.org/{row['geonames']['value']}/")
        
        if 'grWork' in row:
            batch_ids[qid]['goodreads'].append(f"https://www.goodreads.com/work/{row['grWork']['value']}")

    return batch_ids

# Get creation year
def extract_year(label_lit):
    return int(str(label_lit))

def get_creation_year(g, expr):
    # Expression_Creation
    for ec in g.objects(expr, LRMOO.R17i_was_created_by):
        for ts in g.objects(ec, ECRM["P4_has_time-span"]):
            return extract_year(g.value(ts, RDFS.label))
    # Manifestation_Creation fallback
    for manif in g.objects(expr, LRMOO.R4i_is_embodied_in):
        for mc in g.subjects(LRMOO.R24_created, manif):
            for ts in g.objects(mc, ECRM["P4_has_time-span"]):
                return extract_year(g.value(ts, RDFS.label))
    return None

# Arguments
def parse_args(argv=None):
    p = argparse.ArgumentParser(
        description="Add ontology alignments and external identifiers (DBpedia, GeoNames, …) to a TTL file."
    )
    p.add_argument("--input",  type=Path, help="Input TTL (e.g. examples/outputs/all.ttl)")
    p.add_argument("--output", type=Path, help="Output TTL (default: <input>_mapped-and-aligned.ttl)")
    p.add_argument("--log-level", default="INFO", help="Logging level (default: INFO)")
    return p.parse_args(argv)

# Run script
def main(argv=None):
    args = parse_args(argv)

    import logging
    logging.basicConfig(level=getattr(logging, args.log_level.upper(), logging.INFO),
                        format="%(levelname)s:%(name)s:%(message)s")

    if args.input is None:
        candidates = [
            Path("examples/outputs/all.ttl"),
            Path("examples/outputs/authors.ttl"),
            Path("examples/outputs/works.ttl"),
            Path("examples/outputs/relations.ttl"),
        ]
        for c in candidates:
            if c.exists():
                args.input = c
                break
        if args.input is None:
            raise SystemExit("--input is required (no default TTL found in examples/outputs)")

    if args.output is None:
        args.output = args.input.with_name(args.input.stem + "_mapped-and-aligned.ttl")

    args.output.parent.mkdir(parents=True, exist_ok=True)

    g = Graph()
    g.parse(str(args.input), format="turtle")
    print(f"✅ Loaded {args.input}.")

    # Mapping
    
    prefix_map = {
        "schema":  "https://schema.org/",
        "dbpedia": "https://dbpedia.org/"
    }

    wikidata_pattern = re.compile(r"^http://www\.wikidata\.org/entity/(Q\d+)$")
    subjects_by_qid = {}
    for subj, _, obj in g.triples((None, OWL.sameAs, None)):
        m = wikidata_pattern.match(str(obj))
        if m:
            subjects_by_qid.setdefault(m.group(1), []).append(subj)

    qids = list(subjects_by_qid.keys())
    batch_results = query_wikidata_batch(qids) if qids else {}

    for qid, subjects in subjects_by_qid.items():
        for uri_list in batch_results.get(qid, {}).values():
            for raw in uri_list:
                uri = normalize_uri(raw, prefix_map)
                if uri.startswith("http://") or uri.startswith("https://"):
                    new_obj = URIRef(uri)
                    for subj in subjects:
                        if (subj, OWL.sameAs, new_obj) not in g:
                            g.add((subj, OWL.sameAs, new_obj))
    
    # Alignment

    g.bind("sappho_prop", SAPPHO_PROP)
    g.bind("ecrm", ECRM)
    g.bind("crm", CRM)
    g.bind("frbroo", FRBROO)
    g.bind("efrbroo", EFRBROO)
    g.bind("lrmoo", LRMOO)
    g.bind("intro", INTRO)
    g.bind("skos", SKOS)
    
    g.bind("bibo", BIBO)
    g.bind("cito", CITO)
    g.bind("dc", DC)
    g.bind("doco", DOCO)
    g.bind("dracor", DRACOR)
    g.bind("fabio", FABIO)
    g.bind("foaf", FOAF)
    g.bind("golem", GOLEM)
    g.bind("intertext_ab", INTERTEXT_AB)
    g.bind("intertext_tx", INTERTEXT_TX)
    g.bind("intertext_af", INTERTEXT_AF)
    g.bind("intertext_mt", INTERTEXT_MT)
    g.bind("mimotext", MIMOTEXT)
    g.bind("ontopoetry_core", ONTOPOETRY_CORE)
    g.bind("ontopoetry_analysis", ONTOPOETRY_ANALYSIS)
    g.bind("schema", SCHEMA)
    g.bind("urw", URW, override=True)
    g.bind("urb", URB, override=True)

    ## Classes ##
    
    # ecrm:E21_Person
    if any(g.triples((None, RDF.type, ECRM.E21_Person))):
        g.add((DRACOR.author, SKOS.broadMatch, ECRM.E21_Person))
        g.add((ECRM.E21_Person, SKOS.broadMatch, FOAF.Agent))
        g.add((MIMOTEXT.Q11, SKOS.broadMatch, ECRM.E21_Person)) # author
        g.add((MIMOTEXT.Q10, SKOS.closeMatch, ECRM.E21_Person)) # person
        g.add((ONTOPOETRY_CORE.Person, SKOS.closeMatch, ECRM.E21_Person)) 
        g.add((URW.Agent, SKOS.narrowMatch, ECRM.E21_Person))
        g.add((URW.Person, SKOS.closeMatch, ECRM.E21_Person))
    
    # ecrm:E35_Title
    if any(g.triples((None, RDF.type, ECRM.E35_Title))):
        g.add((DOCO.Title, SKOS.closeMatch, ECRM.E35_Title))
    
    # ecrm:E38_Image
    if any(g.triples((None, RDF.type, ECRM.E38_Image))):
        g.add((BIBO.Image, SKOS.closeMatch, ECRM.E38_Image))
        g.add((FOAF.Image, SKOS.closeMatch, ECRM.E38_Image))
    
    # ecrm:E40_Legal_Body
    if any(g.triples((None, RDF.type, ECRM.E40_Legal_Body))):
        g.add((ECRM.E40_Legal_Body, SKOS.broadMatch, FOAF.Agent))
        g.add((ONTOPOETRY_CORE.Organisation, SKOS.broadMatch, ECRM.E40_Legal_Body))
        g.add((URW.Organization, SKOS.broadMatch, ECRM.E40_Legal_Body))
        g.add((URW.Publisher, SKOS.broadMatch, ECRM.E40_Legal_Body))

    # ecrm:E52_Time-Span
    if any(g.triples((None, RDF.type, ECRM["E52_Time-Span"]))):
        g.add((DC.PeriodOfTime, SKOS.closeMatch, ECRM["E52_Time-Span"]))
        g.add((ONTOPOETRY_CORE.TimeSpan, SKOS.closeMatch, ECRM["E52_Time-Span"]))

    # ecrm:E53_Place
    if any(g.triples((None, RDF.type, ECRM.E53_Place))):
        g.add((DC.Location, SKOS.closeMatch, ECRM.E53_Place))
        g.add((MIMOTEXT.Q26, SKOS.closeMatch, ECRM.E53_Place)) # spatial concept
        g.add((ONTOPOETRY_CORE.Place, SKOS.closeMatch, ECRM.E53_Place))
        g.add((ONTOPOETRY_CORE.Place, SKOS.closeMatch, ECRM.E53_Place))
        g.add((URW.Place, SKOS.closeMatch, ECRM.E53_Place))

    # ecrm:E55_Type
    if any(g.triples((None, RDF.type, ECRM.E55_Type))):
        g.add((DRACOR.genre, SKOS.broadMatch, ECRM.E55_Type))
        g.add((INTERTEXT_TX.TextGenre, SKOS.broadMatch, ECRM.E55_Type))
        g.add((MIMOTEXT.Q33, SKOS.broadMatch, ECRM.E55_Type)) # genre

    # ecrm:E67_Birth
    if any(g.triples((None, RDF.type, ECRM.E67_Birth))):
        g.add((ONTOPOETRY_CORE.Birth, SKOS.closeMatch, ECRM.E67_Birth))
    
    # ecrm:E69_Death
    if any(g.triples((None, RDF.type, ECRM.E69_Death))):
        g.add((ONTOPOETRY_CORE.Death, SKOS.closeMatch, ECRM.E69_Death))
        
    # ecrm:E73_Information_Object
    if any(g.triples((None, RDF.type, ECRM.E73_Information_Object))):
        g.add((FABIO.DigitalItem, SKOS.broadMatch, ECRM.E73_Information_Object))  
    
    # lrmoo:F1_Work
    if any(g.triples((None, RDF.type, LRMOO.F1_Work))):
        g.add((FABIO.Work, SKOS.closeMatch, LRMOO.F1_Work))
        g.add((FABIO.LiteraryArtisticWork, SKOS.broadMatch, LRMOO.F1_Work))
        g.add((ONTOPOETRY_CORE.Work, SKOS.closeMatch, LRMOO.F1_Work))
        g.add((ONTOPOETRY_CORE.PoeticWork, SKOS.broadMatch, LRMOO.F1_Work))
        g.add((URB.Work, SKOS.closeMatch, LRMOO.F1_Work))
    
    # lrmoo:F2_Expression
    if any(g.triples((None, RDF.type, LRMOO.F2_Expression))):
        g.add((FOAF.Document, SKOS.broadMatch, LRMOO.F2_Expression))
        g.add((BIBO.Manuscript, SKOS.broadMatch, LRMOO.F2_Expression))
        g.add((DRACOR.play, SKOS.broadMatch, LRMOO.F2_Expression))
        g.add((FABIO.Expression, SKOS.closeMatch, LRMOO.F2_Expression))
        g.add((INTERTEXT_TX.Text, SKOS.broadMatch, LRMOO.F2_Expression))
        g.add((INTERTEXT_TX.SingleText, SKOS.broadMatch, LRMOO.F2_Expression))
        g.add((INTERTEXT_AF.Work, SKOS.broadMatch, LRMOO.F2_Expression))
        g.add((INTERTEXT_AB.Reference, SKOS.broadMatch, LRMOO.F2_Expression))
        g.add((MIMOTEXT.Q2, SKOS.broadMatch, LRMOO.F2_Expression)) # literary work
        g.add((ONTOPOETRY_CORE.Expression, SKOS.closeMatch, LRMOO.F2_Expression))
        g.add((ONTOPOETRY_ANALYSIS.Intertextuality, SKOS.broadMatch, LRMOO.F2_Expression))
        g.add((ONTOPOETRY_ANALYSIS.Source, SKOS.broadMatch, LRMOO.F2_Expression))
        g.add((ONTOPOETRY_ANALYSIS.Redaction, SKOS.broadMatch, LRMOO.F2_Expression))
        g.add((ONTOPOETRY_ANALYSIS.Excerpt, SKOS.broadMatch, LRMOO.F2_Expression))
        g.add((URB.Expression, SKOS.closeMatch, LRMOO.F2_Expression))

    # lrmoo:F3_Manifestation
    if any(g.triples((None, RDF.type, LRMOO.F3_Manifestation))):
        g.add((BIBO.Book, SKOS.broadMatch, LRMOO.F3_Manifestation))
        g.add((DC.BibliographicResource, SKOS.broadMatch, LRMOO.F3_Manifestation))
        g.add((FABIO.Manifestation, SKOS.closeMatch, LRMOO.F3_Manifestation))
        g.add((FOAF.Document, SKOS.broadMatch, LRMOO.F3_Manifestation))
        g.add((URB.Manifestation, SKOS.closeMatch, LRMOO.F3_Manifestation))

    # lrmoo:F5_Item
    if any(g.triples((None, RDF.type, LRMOO.F5_Item))):
        g.add((FABIO.Item, SKOS.closeMatch, LRMOO.F5_Item))
        g.add((FOAF.Document, SKOS.narrowMatch, LRMOO.F5_Item))
    
    # lrmoo:F27_Work_Creation
    if any(g.triples((None, RDF.type, LRMOO.F27_Work_Creation))):
        g.add((ONTOPOETRY_CORE.WorkConception, SKOS.closeMatch, LRMOO.F27_Work_Creation))

    # lrmoo:F28_Expression_Creation
    if any(g.triples((None, RDF.type, LRMOO.F28_Expression_Creation))):
        g.add((ONTOPOETRY_CORE.ExpressionCreation, SKOS.closeMatch, LRMOO.F28_Expression_Creation))
    
    # intro:INT1_Segment
    if any(g.triples((None, RDF.type, INTRO.INT1_Segment))):
        g.add((INTERTEXT_AF.Segment, SKOS.broadMatch, INTRO.INT1_Segment))
    
    # intro:INT2_ActualizationOfFeature
    if any(g.triples((None, RDF.type, INTRO.INT2_ActualizationOfFeature))):
        g.add((FRBROO.F38_Character, SKOS.broadMatch, INTRO.INT2_ActualizationOfFeature))
        g.add((EFRBROO.F38_Character, SKOS.broadMatch, INTRO.INT2_ActualizationOfFeature))
        g.add((DRACOR.character, SKOS.broadMatch, INTRO.INT2_ActualizationOfFeature))
        g.add((GOLEM.G1_Character, SKOS.broadMatch, INTRO.INT2_ActualizationOfFeature))
        g.add((GOLEM.G7_Narrative_Sequence, SKOS.broadMatch, INTRO.INT2_ActualizationOfFeature))
        g.add((ONTOPOETRY_CORE.Character, SKOS.broadMatch, INTRO.INT2_ActualizationOfFeature))
    
    # intro:INT4_Feature
    if any(g.triples((None, RDF.type, INTRO.INT4_Feature))):
        g.add((INTRO.INT4_Feature, SKOS.broadMatch, INTERTEXT_AB.Mediator))
        g.add((GOLEM.G9_Narrative_Unit, SKOS.broadMatch, INTRO.INT4_Feature))
    
    # intro:INT6_Architext
    if any(g.triples((None, RDF.type, INTRO.INT6_Architext))):
        g.add((INTERTEXT_AF.System, SKOS.broadMatch, INTRO.INT6_Architext))
    
    # intro:INT11_TypeOfInterrelation
    if any(g.triples((None, RDF.type, INTRO.INT11_TypeOfInterrelation))):
        g.add((INTERTEXT_AB.IntertexualSpecification, SKOS.closeMatch, INTRO.INT11_TypeOfInterrelation))
    
    # intro:INT21_TextPassage
    if any(g.triples((None, RDF.type, INTRO.INT21_TextPassage))):
        g.add((INTRO.INT21_TextPassage, SKOS.broadMatch, DOCO.Part))
        g.add((INTRO.INT21_TextPassage, SKOS.broadMatch, DOCO.BackMatter))
        g.add((INTRO.INT21_TextPassage, SKOS.broadMatch, DOCO.BodyMatter))
        g.add((INTRO.INT21_TextPassage, SKOS.broadMatch, DOCO.CaptionedBox))
        g.add((INTRO.INT21_TextPassage, SKOS.broadMatch, DOCO.Chapter))
        g.add((INTRO.INT21_TextPassage, SKOS.broadMatch, DOCO.ComplexRunInQuotation))
        g.add((INTRO.INT21_TextPassage, SKOS.broadMatch, DOCO.Footnote))
        g.add((INTRO.INT21_TextPassage, SKOS.broadMatch, DOCO.Formula))
        g.add((INTRO.INT21_TextPassage, SKOS.broadMatch, DOCO.FormulaBox))
        g.add((INTRO.INT21_TextPassage, SKOS.broadMatch, DOCO.FrontMatter))
        g.add((INTRO.INT21_TextPassage, SKOS.broadMatch, DOCO.List))
        g.add((INTRO.INT21_TextPassage, SKOS.broadMatch, DOCO.Section))
        g.add((INTRO.INT21_TextPassage, SKOS.broadMatch, DOCO.Table))
        g.add((INTRO.INT21_TextPassage, SKOS.broadMatch, INTERTEXT_AB.Mediator))
        g.add((BIBO.Quote, SKOS.broadMatch, INTRO.INT21_TextPassage))
        g.add((FABIO.Quotation, SKOS.broadMatch, INTRO.INT21_TextPassage))
        g.add((INTERTEXT_TX.TextSegment, SKOS.closeMatch, INTRO.INT21_TextPassage))
    
    # intro:INT31_IntertextualRelation
    if any(g.triples((None, RDF.type, INTRO.INT31_IntertextualRelation))):
        g.add((INTERTEXT_AB.IntertexualRelation, SKOS.closeMatch, INTRO.INT31_IntertextualRelation))
        g.add((URW.EntityInfluence, SKOS.narrowMatch, INTRO.INT31_IntertextualRelation))
        g.add((URB.Reception, SKOS.narrowMatch, INTRO.INT31_IntertextualRelation))

    # intro:INT_Character
    if any(g.triples((None, RDF.type, INTRO.INT_Character))):
        g.add((GOLEM["G0_Character-Stoff"], SKOS.closeMatch, INTRO.INT_Character))
        g.add((FRBROO.F38_Character, SKOS.broadMatch, INTRO.INT_Character))
        g.add((EFRBROO.F38_Character, SKOS.broadMatch, INTRO.INT_Character))
        g.add((DRACOR.character, SKOS.broadMatch, INTRO.INT_Character))
        g.add((ONTOPOETRY_CORE.Character, SKOS.broadMatch, INTRO.INT_Character))
    
    # intro:INT_Plot
    if any(g.triples((None, RDF.type, INTRO.INT_Plot))):
        g.add((GOLEM.G14_Narrative_Stoff, SKOS.closeMatch, INTRO.INT_Plot))
    
    # intro:INT_Motif
    if any(g.triples((None, RDF.type, INTRO.INT_Motif))):
        g.add((INTRO.INT_Motif, SKOS.broadMatch, INTERTEXT_MT.Motive))
    
    # intro:INT_Topic
    if any(g.triples((None, RDF.type, INTRO.INT_Topic))):
        g.add((MIMOTEXT.Q20, SKOS.closeMatch, INTRO.INT_Topic)) # thematic concept
    
    ## Properties ##
    
    # ecrm:P1_is_identified_by
    if any(g.triples((None, ECRM.P1_is_identified_by, None))):
        g.add((DC.identifier, SKOS.closeMatch, ECRM.P1_is_identified_by))
        g.add((URW.hasIdentifier, SKOS.closeMatch, ECRM.P1_is_identified_by))

    # ecrm:P2_has_type
    if any(g.triples((None, ECRM.P2_has_type, None))):
        g.add((DC.type, SKOS.closeMatch, ECRM.P2_has_type))
        g.add((DRACOR.has_genre, SKOS.broadMatch, ECRM.P2_has_type))
        g.add((FOAF.gender, SKOS.broadMatch, ECRM.P2_has_type))
        g.add((MIMOTEXT.P12, SKOS.broadMatch, ECRM.P2_has_type)) # genre
        g.add((ONTOPOETRY_CORE.gender, SKOS.broadMatch, ECRM.P2_has_type))
        g.add((ONTOPOETRY_CORE.genre, SKOS.broadMatch, ECRM.P2_has_type))
        g.add((SCHEMA.genre, SKOS.broadMatch, ECRM.P2_has_type))
        g.add((URW.gender, SKOS.broadMatch, ECRM.P2_has_type))
    
    # ecrm:P4_has_time-span
    if any(g.triples((None, ECRM["P4_has_time-span"], None))):
        g.add((DC.date, SKOS.closeMatch, ECRM["P4_has_time-span"]))
        g.add((DC.created, SKOS.broadMatch, ECRM["P4_has_time-span"]))
        g.add((DC.dateCopyrighted, SKOS.broadMatch, ECRM["P4_has_time-span"]))
        g.add((DRACOR.printYear, SKOS.broadMatch, ECRM["P4_has_time-span"]))
        g.add((DRACOR.writtenYear, SKOS.broadMatch, ECRM["P4_has_time-span"]))
        g.add((MIMOTEXT.P9, SKOS.broadMatch, ECRM["P4_has_time-span"])) # publication date
        g.add((ONTOPOETRY_CORE.hasTimeSpan, SKOS.closeMatch, ECRM["P4_has_time-span"]))
        g.add((SCHEMA.dateCreated, SKOS.broadMatch, ECRM["P4_has_time-span"]))
        g.add((SCHEMA.datePublished, SKOS.broadMatch, ECRM["P4_has_time-span"]))
        g.add((URW.wasPublishedWhen, SKOS.broadMatch, ECRM["P4_has_time-span"]))
        g.add((URB.date, SKOS.closeMatch, ECRM["P4_has_time-span"]))
    
    # ecrm:P4i_is_time-span_of
    if any(g.triples((None, ECRM["P4i_is_time-span_of"], None))):
        g.add((ONTOPOETRY_CORE.isTimeSpanOf, SKOS.closeMatch, ECRM["P4i_is_time-span_of"]))
    
    # ecrm:P7_took_place_at
    if any(g.triples((None, ECRM.P7_took_place_at, None))):
        g.add((FABIO.hasPlaceOfPublication, SKOS.broadMatch, ECRM.P7_took_place_at))
        g.add((MIMOTEXT.P10, SKOS.broadMatch, ECRM.P7_took_place_at)) # publication place
        g.add((ONTOPOETRY_CORE.tookPlaceAt, SKOS.closeMatch, ECRM.P7_took_place_at))
        g.add((SCHEMA.locationCreated, SKOS.broadMatch, ECRM.P7_took_place_at))
        g.add((URW.wasPublishedWhere, SKOS.broadMatch, ECRM.P7_took_place_at))
    
    # ecrm:P7i_witnessed
    if any(g.triples((None, ECRM.P7i_witnessed, None))):
        g.add((ONTOPOETRY_CORE.witnessed, SKOS.closeMatch, ECRM.P7i_witnessed))
    
    # ecrm:P14_carried_out_by
    if any(g.triples((None, ECRM.P14_carried_out_by, None))):
        g.add((BIBO.editor, SKOS.broadMatch, ECRM.P14_carried_out_by))
        g.add((DRACOR.has_author, SKOS.broadMatch, ECRM.P14_carried_out_by))
        g.add((FOAF.maker, SKOS.broadMatch, ECRM.P14_carried_out_by))
        g.add((MIMOTEXT.P5, SKOS.broadMatch, ECRM.P14_carried_out_by)) # has author
        g.add((SCHEMA.author, SKOS.broadMatch, ECRM.P14_carried_out_by))
        g.add((SCHEMA.creator, SKOS.broadMatch, ECRM.P14_carried_out_by))
        g.add((URW.wasPublishedBy, SKOS.broadMatch, ECRM.P14_carried_out_by))

    # ecrm:P14i_performed
    if any(g.triples((None, ECRM.P14i_performed, None))):
        g.add((DC.creator, SKOS.broadMatch, ECRM.P14i_performed))
        g.add((DC.publisher, SKOS.broadMatch, ECRM.P14i_performed))
        g.add((FOAF.made, SKOS.broadMatch, ECRM.P14i_performed))
        g.add((MIMOTEXT.P7, SKOS.broadMatch, ECRM.P14i_performed)) # author of
    
    # ecrm:P98i_was_born
    if any(g.triples((None, ECRM.P98i_was_born, None))):
        g.add((ONTOPOETRY_CORE.wasBorn, SKOS.closeMatch, ECRM.P98i_was_born))
    
    # ecrm:P98_brought_into_life
    if any(g.triples((None, ECRM.P98_brought_into_life, None))):
        g.add((ONTOPOETRY_CORE.broughtIntoLife, SKOS.closeMatch, ECRM.P98_brought_into_life))
    
    # ecrm:P100_was_death_of
    if any(g.triples((None, ECRM.P100_was_death_of, None))):
        g.add((ONTOPOETRY_CORE.wasDeathOf, SKOS.closeMatch, ECRM.P100_was_death_of))
    
    # ecrm:P100i_died_in
    if any(g.triples((None, ECRM.P100i_died_in, None))):
        g.add((ONTOPOETRY_CORE.diedIn, SKOS.closeMatch, ECRM.P100i_died_in))
        
    # ecrm:P102_has_title
    if any(g.triples((None, ECRM.P102_has_title, None))):
        g.add((DC.title, SKOS.closeMatch, ECRM.P102_has_title))
        g.add((MIMOTEXT.P4, SKOS.closeMatch, ECRM.P102_has_title))  # title
    
    # ecrm:P131_is_identified_by
    if any(g.triples((None, ECRM.P131_is_identified_by, None))):
        g.add((FOAF.name, SKOS.closeMatch, ECRM.P131_is_identified_by))
        g.add((MIMOTEXT.P8, SKOS.closeMatch, ECRM.P131_is_identified_by))  # name
    
    # ecrm:P138i_has_representation
    if any(g.triples((None, ECRM.P138i_has_representation, None))):
        g.add((FOAF.img, SKOS.broadMatch, ECRM.P138i_has_representation))
        g.add((MIMOTEXT.P21, SKOS.broadMatch, ECRM.P138i_has_representation)) # full work available at URL
    
    # lrmoo:R3_realises
    if any(g.triples((None, LRMOO.R3_realises, None))):
        g.add((URB.realization, SKOS.closeMatch, LRMOO.R3_realises))

    # lrmoo:R3_is_realised_in
    if any(g.triples((None, LRMOO.R3_is_realised_in, None))):
        g.add((URB.realizationOf, SKOS.closeMatch, LRMOO.R3_is_realised_in))
    
    # lrmoo:R4_embodies
    if any(g.triples((None, LRMOO.R4_embodies, None))):
        g.add((URB.embodimentOf, SKOS.closeMatch, LRMOO.R4_embodies))

    # lrmoo:R4i_is_embodied_in
    if any(g.triples((None, LRMOO.R4i_is_embodied_in, None))):
        g.add((URB.embodiment, SKOS.closeMatch, LRMOO.R4i_is_embodied_in))
    
    # lrmoo:R16_created
    if any(g.triples((None, LRMOO.R16_created, None))):
        g.add((ONTOPOETRY_CORE.initiated, SKOS.closeMatch, LRMOO.R16_created))
    
    # lrmoo:R16i_was_created_by
    if any(g.triples((None, LRMOO.R16i_was_created_by, None))):
        g.add((ONTOPOETRY_CORE.wasInitiatedBy, SKOS.closeMatch, LRMOO.R16i_was_created_by))
    
    # lrmoo:R17_created
    if any(g.triples((None, LRMOO.R17_created, None))):
        g.add((ONTOPOETRY_CORE.createdExpressionFromExpressionCreation, SKOS.closeMatch, LRMOO.R17_created))
    
    # lrmoo:R17i_was_created_by
    if any(g.triples((None, LRMOO.R17i_was_created_by, None))):
        g.add((ONTOPOETRY_CORE.wasCreatedByExpressionCreationForExpression, SKOS.closeMatch, LRMOO.R17i_was_created_by))
    
    # lrmoo:R19_created_a_realisation_of
    if any(g.triples((None, LRMOO.R19_created_a_realisation_of, None))):
        g.add((ONTOPOETRY_CORE.createdWorkByExpressionCreation, SKOS.closeMatch, LRMOO.R19_created_a_realisation_of))
    
    # lrmoo:R19i_was_realised_through
    if any(g.triples((None, LRMOO.R19i_was_realised_through, None))):
        g.add((ONTOPOETRY_CORE.realisedThroughExpressionCreation, SKOS.closeMatch, LRMOO.R19i_was_realised_through))
    
    # intro:R12i_isReferredToEntity
    if any(g.triples((None, INTRO.R12i_isReferredToEntity, None))):
        g.add((INTRO.R12i_isReferredToEntity, SKOS.closeMatch, INTERTEXT_AB.there))
    
    # intro:R13i_isReferringEntity
    if any(g.triples((None, INTRO.R13i_isReferringEntity, None))):
        g.add((INTRO.R13i_isReferringEntity, SKOS.closeMatch, INTERTEXT_AB.here))
    
    # intro:R18_showsActualization
    if any(g.triples((None, INTRO.R18_showsActualization, None))):
        g.add((ONTOPOETRY_ANALYSIS.presentsIntertextuality, SKOS.broadMatch, INTRO.R18_showsActualization))
    
    # intro:R18i_actualizationFoundOn
    if any(g.triples((None, INTRO.R18i_actualizationFoundOn, None))):
        g.add((ONTOPOETRY_ANALYSIS.isIntertextualityPresentAt, SKOS.broadMatch, INTRO.R18i_actualizationFoundOn))

    # intro:R19i_isTypeOf
    if any(g.triples((None, INTRO.R19i_isTypeOf, None))):
        g.add((INTERTEXT_AB.specifiedBy, SKOS.broadMatch, INTRO.R19i_isTypeOf))
        g.add((ONTOPOETRY_ANALYSIS.typeOfIntertextuality, SKOS.broadMatch, INTRO.R19i_isTypeOf))
    
    # intro:R22i_relationIsBasedOnSimilarity
    if any(g.triples((None, INTRO.R22i_relationIsBasedOnSimilarity, None))):
        g.add((INTRO.R22i_relationIsBasedOnSimilarity, SKOS.broadMatch, INTERTEXT_AB.mediatedBy))
    
    # intro:R24_hasRelatedEntity
    if any(g.triples((None, INTRO.R24_hasRelatedEntity, None))):
        g.add((INTRO.R24_hasRelatedEntity, SKOS.broadMatch, INTERTEXT_AB.mediatedBy))
    
    # intro:R30_hasTextPassage
    if any(g.triples((None, INTRO.R30_hasTextPassage, None))):
        g.add((INTRO.R30_hasTextPassage, SKOS.broadMatch, DC.hasPart))
        g.add((ONTOPOETRY_ANALYSIS.presentsIntertextuality, SKOS.broadMatch, INTRO.R30_hasTextPassage))
    
    # intro:R30i_isTextPassageOf
    if any(g.triples((None, INTRO.R30i_isTextPassageOf, None))):
        g.add((ONTOPOETRY_ANALYSIS.isIntertextualityPresentAt, SKOS.broadMatch, INTRO.R30i_isTextPassageOf))
    
    # prov:wasDerivedFrom
    if any(g.triples((None, PROV.wasDerivedFrom, None))):
        g.add((DC.source, SKOS.closeMatch, PROV.wasDerivedFrom))
        g.add((MIMOTEXT.P17, SKOS.broadMatch, PROV.wasDerivedFrom)) # reference URL

    ## Complex Properties ##    

    # new properties for F1->F3 (hasManifestation), F1->F5 (hasPortrayal) and F2->F5 (hasRepresentation)
    if (
        any(g.triples((None, RDF.type, LRMOO.F1_Work))) and
        any(g.triples((None, RDF.type, LRMOO.F3_Manifestation)))
    ):
        g.add((SAPPHO_PROP.has_manifestation, RDF.type, OWL.ObjectProperty))
        g.add((SAPPHO_PROP.has_manifestation, RDFS.label, 
            Literal("has manifestation", lang="en")))
        g.add((SAPPHO_PROP.has_manifestation, SKOS.closeMatch, FABIO.hasManifestation))
        g.add((SAPPHO_PROP.has_manifestation, SKOS.closeMatch, URB.manifestation))
        g.add((SAPPHO_PROP.has_manifestation, RDFS.domain, LRMOO.F1_Work))
        g.add((SAPPHO_PROP.has_manifestation, RDFS.range,  LRMOO.F3_Manifestation))

        bnode = BNode()
        Collection(g, bnode, [
            LRMOO.R3_is_realised_in,
            LRMOO.R4i_is_embodied_in
        ])
        g.add((SAPPHO_PROP.has_manifestation, OWL.propertyChainAxiom, bnode))
        
        for work in g.subjects(RDF.type, LRMOO.F1_Work):
            for expr in g.objects(work, LRMOO.R3_is_realised_in):
                for mani in g.objects(expr, LRMOO.R4i_is_embodied_in):
                    g.add((work, SAPPHO_PROP.has_manifestation, mani))
  
    if (
        any(g.triples((None, RDF.type, LRMOO.F1_Work))) and
        any(g.triples((None, RDF.type, LRMOO.F5_Item)))
    ):
        g.add((SAPPHO_PROP.has_portrayal, RDF.type, OWL.ObjectProperty))
        g.add((SAPPHO_PROP.has_portrayal, RDFS.label, 
            Literal("has portrayal", lang="en")))
        g.add((SAPPHO_PROP.has_portrayal, SKOS.closeMatch, FABIO.hasPortrayal))
        g.add((SAPPHO_PROP.has_portrayal, RDFS.domain, LRMOO.F1_Work))
        g.add((SAPPHO_PROP.has_portrayal, RDFS.range,  LRMOO.F5_Item))

        bnode = BNode()
        Collection(g, bnode, [
            LRMOO.R3_is_realised_in,
            LRMOO.R4i_is_embodied_in,
            LRMOO.R7i_is_exemplified_by
        ])
        g.add((SAPPHO_PROP.has_portrayal, OWL.propertyChainAxiom, bnode))
        
        for work in g.subjects(RDF.type, LRMOO.F1_Work):
            for expr in g.objects(work, LRMOO.R3_is_realised_in):
                for mani in g.objects(expr, LRMOO.R4i_is_embodied_in):
                    for item in g.objects(mani, LRMOO.R7i_is_exemplified_by):
                        g.add((work, SAPPHO_PROP.has_portrayal, item))

    if (
        any(g.triples((None, RDF.type, LRMOO.F2_Expression))) and
        any(g.triples((None, RDF.type, LRMOO.F5_Item)))
    ):
        g.add((SAPPHO_PROP.has_representation, RDF.type, OWL.ObjectProperty))
        g.add((SAPPHO_PROP.has_representation, RDFS.label, 
            Literal("has representation", lang="en")))
        g.add((SAPPHO_PROP.has_representation, SKOS.closeMatch, FABIO.hasRepresentation))
        g.add((SAPPHO_PROP.has_representation, RDFS.domain, LRMOO.F2_Expression))
        g.add((SAPPHO_PROP.has_representation, RDFS.range,  LRMOO.F5_Item))

        bnode = BNode()
        Collection(g, bnode, [
            LRMOO.R4i_is_embodied_in,
            LRMOO.R7i_is_exemplified_by
        ])
        g.add((SAPPHO_PROP.has_representation, OWL.propertyChainAxiom, bnode))
        
        for expr in g.subjects(RDF.type, LRMOO.F2_Expression):
                for mani in g.objects(expr, LRMOO.R4i_is_embodied_in):
                    for item in g.objects(mani, LRMOO.R7i_is_exemplified_by):
                        g.add((expr, SAPPHO_PROP.has_representation, item))
    
    # sappho_prop:about: expressions that actualize a intro:INT_Topic will be also linked to the topic
    
    # Gather directions (younger/older) first, needed later as well
    directions = []
    
    for rel in g.subjects(RDF.type, INTRO.INT31_IntertextualRelation):
        tp_expr = []
        for tp in g.objects(rel, INTRO.R24_hasRelatedEntity):
            expr = g.value(tp, INTRO.R30i_isTextPassageOf)
            if expr:
                tp_expr.append((tp, expr))

        if len(tp_expr) != 2:
            continue  
        
        exprs = {e for tp, e in tp_expr}
        if len(exprs) != 2:
                continue
                
        (tp1, expr1), (tp2, expr2) = tp_expr[:2]
            
        y1 = get_creation_year(g, expr1)
        y2 = get_creation_year(g, expr2)
        
        if y1 is None or y2 is None:
            continue
        
        if y1 < y2:
            older_expr, younger_expr = expr1, expr2
            older_tp,    younger_tp    = tp1,    tp2
        else:
            older_expr, younger_expr = expr2, expr1
            older_tp,    younger_tp    = tp2,    tp1

        directions.append((younger_expr, older_expr, younger_tp, older_tp))

    if any(g.triples((None, RDF.type, INTRO.INT_Topic))):
        g.add((SAPPHO_PROP.about, RDF.type, OWL.ObjectProperty))
        g.add((SAPPHO_PROP.about, RDFS.label,
            Literal("Link from expression to topic", lang="en")))
        b_about = BNode()
        Collection(g, b_about, [
            INTRO.R18_showsActualization,
            INTRO.R17_actualizesFeature
        ])
        g.add((SAPPHO_PROP.about, OWL.propertyChainAxiom, b_about))
        g.add((SAPPHO_PROP.about, SKOS.closeMatch, DC.subject))
        g.add((SAPPHO_PROP.about, SKOS.closeMatch, FOAF.topic))
        g.add((SAPPHO_PROP.about, SKOS.closeMatch, MIMOTEXT.P36)) # about
        g.add((SAPPHO_PROP.about, SKOS.closeMatch, SCHEMA.about))
        g.add((SAPPHO_PROP.about, RDFS.domain, LRMOO.F2_Expression))
        g.add((SAPPHO_PROP.about, RDFS.range, INTRO.INT_Topic))

        for expr in g.subjects(RDF.type, LRMOO.F2_Expression):
            for act in g.objects(expr, INTRO.R18_showsActualization):
                topic = g.value(act, INTRO.R17_actualizesFeature)
                if topic and (topic, RDF.type, INTRO.INT_Topic) in g:
                    g.add((expr, SAPPHO_PROP.about, topic))

    # sappho_prop:expr_relation: expressions that are linked via intro:INT31 will be linked via this property
    if any(g.triples((None, RDF.type, INTRO.INT31_IntertextualRelation))):
        g.add((SAPPHO_PROP.expr_relation, RDF.type, OWL.ObjectProperty))
        g.add((SAPPHO_PROP.expr_relation, RDFS.label,
            Literal("Relation between two expressions", lang="en")))

        first_elem = BNode()
        g.add((first_elem, OWL.inverseOf, INTRO.R18i_actualizationFoundOn))
        chain_list = [
            first_elem,
            INTRO.R24i_isRelatedEntity,
            INTRO.R24_hasRelatedEntity,
            INTRO.R18i_actualizationFoundOn
        ]
        chain_bnode = BNode()
        Collection(g, chain_bnode, chain_list)
        g.add((SAPPHO_PROP.expr_relation, OWL.propertyChainAxiom, chain_bnode))
        g.add((SAPPHO_PROP.expr_relation, RDF.type, OWL.SymmetricProperty))
        g.add((SAPPHO_PROP.expr_relation, SKOS.closeMatch, DC.relation))
        g.add((SAPPHO_PROP.expr_relation, SKOS.closeMatch, MIMOTEXT.P34))  # relation
        g.add((SAPPHO_PROP.expr_relation, SKOS.narrowMatch, URW.influenced))
        g.add((SAPPHO_PROP.expr_relation, SKOS.narrowMatch, URW.influencedBy))
        g.add((SAPPHO_PROP.expr_relation, RDFS.domain, LRMOO.F2_Expression))
        g.add((SAPPHO_PROP.expr_relation, RDFS.range,  LRMOO.F2_Expression))
        
        # add directions to intro:INT31_IntertextualRelation
        for rel in g.subjects(RDF.type, INTRO.INT31_IntertextualRelation):
            acts = list(g.objects(rel, INTRO.R24_hasRelatedEntity))
            exprs = {
                expr
                for act in acts
                for expr in g.subjects(INTRO.R18_showsActualization, act)
            }
            for e1 in exprs:
                for e2 in exprs:
                    if e1 != e2:
                        g.add((e1, SAPPHO_PROP.expr_relation, e2))
                        g.add((e2, SAPPHO_PROP.expr_relation, e1))

        # Also materialize younger/older hints if computed
        for younger_expr, older_expr, younger_tp, older_tp in directions:
            g.add((URIRef(str(rel)), INTRO.R13_hasReferringEntity, younger_expr))
            g.add((younger_expr, INTRO.R13i_isReferringEntity, URIRef(str(rel))))
            g.add((URIRef(str(rel)), INTRO.R12_hasReferredToEntity, older_expr))
            g.add((older_expr, INTRO.R12i_isReferredToEntity, URIRef(str(rel))))

    # sappho_prop:expr_possibly_cites / sappho_prop:expr_possibly_cited_by
    # if two expressions have intro:INT21_TextPassages that are part of their intro:INT31, 
    # it is possible (but not necessary) that the younger text cites the older text. 
    # To find out which one is which, the time-spans of the expression or manifestation creations are compared.

    if any(g.triples((None, INTRO.R30i_isTextPassageOf, None))):

        g.add((SAPPHO_PROP.expr_possibly_cites, RDF.type, OWL.ObjectProperty))
        g.add((SAPPHO_PROP.expr_possibly_cites, RDFS.label,
            Literal("Younger expression possibly cites older expression", lang="en")))
        g.add((SAPPHO_PROP.expr_possibly_cites, RDFS.domain, LRMOO.F2_Expression))
        g.add((SAPPHO_PROP.expr_possibly_cites, RDFS.range, LRMOO.F2_Expression))

        inv_rel = BNode(); g.add((inv_rel, OWL.inverseOf, INTRO.R24_hasRelatedEntity))
        inv_tp  = BNode()
        g.add((inv_tp, OWL.inverseOf, INTRO.R30i_isTextPassageOf))

        chain_nodes = [
            INTRO.R30_hasTextPassage, 
            inv_rel,                 
            INTRO.R24_hasRelatedEntity,
            inv_tp                 
    ]
        chain_bnode = BNode(); Collection(g, chain_bnode, chain_nodes)
        
        g.add((SAPPHO_PROP.expr_possibly_cites, OWL.propertyChainAxiom, chain_bnode))

        g.add((SAPPHO_PROP.expr_possibly_cited_by, RDF.type, OWL.ObjectProperty))
        g.add((SAPPHO_PROP.expr_possibly_cites, RDFS.label,
            Literal("Older expression possibly cited by younger expression", lang="en")))
        g.add((SAPPHO_PROP.expr_possibly_cited_by, OWL.inverseOf,
            SAPPHO_PROP.expr_possibly_cites))
        g.add((SAPPHO_PROP.expr_possibly_cited_by, RDFS.domain, LRMOO.F2_Expression))
        g.add((SAPPHO_PROP.expr_possibly_cited_by, RDFS.range, LRMOO.F2_Expression))

        # ecrm alignment
        g.add((SAPPHO_PROP.expr_possibly_cites, RDFS.subPropertyOf, ECRM.P148_has_component))
        g.add((SAPPHO_PROP.expr_possibly_cited_by, RDFS.subPropertyOf, ECRM.P148i_is_component_of))   
                
        g.add((SAPPHO_PROP.expr_possibly_cites, RDFS.subPropertyOf, ECRM.P130_shows_features_of))
        g.add((SAPPHO_PROP.expr_possibly_cited_by, RDFS.subPropertyOf, ECRM.P130i_features_are_also_found_on))   
        
        # lrmoo alignment
        g.add((LRMOO.R76_is_derivative_of, SKOS.broadMatch, SAPPHO_PROP.expr_possibly_cites))
        g.add((LRMOO.R76i_has_derivative, SKOS.broadMatch, SAPPHO_PROP.expr_possibly_cited_by))

        # other ontologies
        g.add((BIBO.cites, SKOS.broadMatch, SAPPHO_PROP.expr_possibly_cites))
        g.add((BIBO.citedBy, SKOS.broadMatch, SAPPHO_PROP.expr_possibly_cited_by))
        
        g.add((CITO.cites, SKOS.broadMatch, SAPPHO_PROP.expr_possibly_cites))
        g.add((CITO.isCitedBy, SKOS.broadMatch, SAPPHO_PROP.expr_possibly_cited_by))
        
        g.add((ONTOPOETRY_ANALYSIS.usedAsRedaction, SKOS.broadMatch, SAPPHO_PROP.expr_possibly_cites)) # young to old
        g.add((ONTOPOETRY_ANALYSIS.usedAsSource, SKOS.broadMatch, SAPPHO_PROP.expr_possibly_cites)) # young to old
        g.add((ONTOPOETRY_ANALYSIS.showsInfluencesOf, SKOS.broadMatch, SAPPHO_PROP.expr_possibly_cites)) # young to old
        g.add((ONTOPOETRY_ANALYSIS.isDerivativeOf, SKOS.broadMatch, SAPPHO_PROP.expr_possibly_cites)) # young to old
        g.add((ONTOPOETRY_ANALYSIS.isUsedRedactionIn, SKOS.broadMatch, SAPPHO_PROP.expr_possibly_cited_by)) # old to young
        g.add((ONTOPOETRY_ANALYSIS.isUsedSourceIn, SKOS.broadMatch, SAPPHO_PROP.expr_possibly_cited_by)) # old to young
        g.add((ONTOPOETRY_ANALYSIS.influencesAreFoundOn, SKOS.broadMatch, SAPPHO_PROP.expr_possibly_cited_by)) # old to young
        g.add((ONTOPOETRY_ANALYSIS.hasDerivative, SKOS.broadMatch, SAPPHO_PROP.expr_possibly_cited_by)) # old to young
        
        g.add((SCHEMA.citation, SKOS.broadMatch, SAPPHO_PROP.expr_possibly_cites))
        
        # sappho_prop:tp_possibly_cites / sappho_prop:tp_possibly_cited_by
        # if two expressions have intro:INT21_TextPassages that are part of their intro:INT31, 
        # it is also possible (but not necessary) that the younger text cites the text passage of the older text. 
                            
        g.add((SAPPHO_PROP.tp_possibly_cites, RDF.type, OWL.ObjectProperty))
        g.add((SAPPHO_PROP.tp_possibly_cites, RDFS.label,
            Literal("The younger text possibly cites the text passage of the older text", lang="en")))
        g.add((SAPPHO_PROP.tp_possibly_cites, RDFS.domain, LRMOO.F2_Expression))
        g.add((SAPPHO_PROP.tp_possibly_cites, RDFS.range, INTRO.INT21_TextPassage))
        
        g.add((SAPPHO_PROP.tp_possibly_cited_by, RDF.type, OWL.ObjectProperty))
        g.add((SAPPHO_PROP.tp_possibly_cited_by, RDFS.label,
            Literal("The older text is possibly cited by the younger text passage", lang="en")))
        g.add((SAPPHO_PROP.tp_possibly_cited_by, RDFS.domain, INTRO.INT21_TextPassage))
        g.add((SAPPHO_PROP.tp_possibly_cited_by, RDFS.range, LRMOO.F2_Expression))
        
        g.add((SAPPHO_PROP.tp_possibly_cited_by, OWL.inverseOf, SAPPHO_PROP.tp_possibly_cites))
        
        g.add((LRMOO.R75_incorporates,  SKOS.broadMatch, SAPPHO_PROP.tp_possibly_cites))
        g.add((LRMOO.R75i_is_incorporated_in, SKOS.broadMatch, SAPPHO_PROP.tp_possibly_cites))
        
        g.add((SAPPHO_PROP.tp_possibly_cites, RDFS.subPropertyOf, ECRM.P148_has_component))
        g.add((SAPPHO_PROP.tp_possibly_cited_by, RDFS.subPropertyOf, ECRM.P148i_is_component_of))   
        
        g.add((SAPPHO_PROP.tp_possibly_cites, RDFS.subPropertyOf, ECRM.P130_shows_features_of))
        g.add((SAPPHO_PROP.tp_possibly_cited_by, RDFS.subPropertyOf, ECRM.P130i_features_are_also_found_on)) 
        
        chain21 = [INTRO.R30_hasTextPassage]
        b21     = BNode()
        Collection(g, b21, chain21)
        g.add((SAPPHO_PROP.tp_possibly_cites, OWL.propertyChainAxiom, b21))

        for younger_expr, older_expr, younger_tp, older_tp in directions:
            
            g.add((younger_expr, SAPPHO_PROP.expr_possibly_cites, older_expr))
            g.add((older_expr,   SAPPHO_PROP.expr_possibly_cited_by, younger_expr))
            
            g.add((younger_expr, SAPPHO_PROP.tp_possibly_cites, older_tp))
            g.add((older_tp,   SAPPHO_PROP.tp_possibly_cited_by, younger_expr))

        g.add((CITO.hasCitedEntity,  SKOS.broadMatch, SAPPHO_PROP.tp_possibly_cites))
        g.add((CITO.hasCitingEntity, SKOS.broadMatch, SAPPHO_PROP.tp_possibly_cited_by))
    
    # sappho_prop:expr_references / sappho_prop:referenced_by_expr: expressions that actualize a intro:INT18_Reference will be also linked to the referred entity
    if any(g.triples((None, ECRM.P67_refers_to, None))):
        g.add((SAPPHO_PROP.expr_references, RDF.type, OWL.ObjectProperty))
        g.add((SAPPHO_PROP.expr_references, RDFS.label,
            Literal("Reference from expression to person, place or expression", lang="en")))
        chain_bnode = BNode()
        Collection(g, chain_bnode, [
            INTRO.R18_showsActualization,
            ECRM.P67_refers_to
        ])
        g.add((SAPPHO_PROP.expr_references, OWL.propertyChainAxiom, chain_bnode))
        
        g.add((SAPPHO_PROP.expr_references, RDFS.domain, LRMOO.F2_Expression))
        g.add((SAPPHO_PROP.expr_references, RDFS.range, ECRM.E21_Person))
        g.add((SAPPHO_PROP.expr_references, RDFS.range, ECRM.E53_Place))
        g.add((SAPPHO_PROP.expr_references, RDFS.range, LRMOO.F2_Expression))
        g.add((SAPPHO_PROP.expr_references, RDFS.subPropertyOf, ECRM.P67_refers_to))
        
        g.add((SAPPHO_PROP.referenced_by_expr, RDF.type, OWL.ObjectProperty))
        g.add((SAPPHO_PROP.referenced_by_expr, RDFS.label,
            Literal("Person, place or expression referenced by expression", lang="en")))
        g.add((SAPPHO_PROP.referenced_by_expr, OWL.inverseOf, SAPPHO_PROP.expr_references))
        g.add((SAPPHO_PROP.referenced_by_expr, RDFS.domain, ECRM.E21_Person))
        g.add((SAPPHO_PROP.referenced_by_expr, RDFS.domain, ECRM.E53_Place))
        g.add((SAPPHO_PROP.referenced_by_expr, RDFS.domain, LRMOO.F2_Expression))
        g.add((SAPPHO_PROP.referenced_by_expr, RDFS.range,  LRMOO.F2_Expression))
        g.add((SAPPHO_PROP.referenced_by_expr, RDFS.subPropertyOf, ECRM.P67i_is_referred_to_by))
        
        for expr in g.subjects(RDF.type, LRMOO.F2_Expression):
            for act in g.objects(expr, INTRO.R18_showsActualization):
                for target in g.objects(act, ECRM.P67_refers_to):
                    g.add((expr, SAPPHO_PROP.expr_references, target))
                    if (target, RDF.type, ECRM.E21_Person) in g:
                        g.add((expr, SAPPHO_PROP.references_person, target))
                    elif (target, RDF.type, ECRM.E53_Place) in g:
                        g.add((expr, SAPPHO_PROP.references_place, target))
                    elif (target, RDF.type, LRMOO.F2_Expression) in g:
                        pass
        
        g.add((SAPPHO_PROP.expr_references, SKOS.closeMatch, DC.references))
        g.add((DC.isReferencedBy, OWL.inverseOf, DC.references))
        
        g.add((SAPPHO_PROP.expr_references, SKOS.narrowMatch, MIMOTEXT.P50)) # mentions
        g.add((MIMOTEXT.P51, OWL.inverseOf, MIMOTEXT.P50))

        g.add((ONTOPOETRY_CORE.mentions, SKOS.broadMatch, SAPPHO_PROP.expr_references))
        g.add((ONTOPOETRY_CORE.isMentionedIn, OWL.inverseOf, ONTOPOETRY_CORE.mentions))
        
        g.add((SCHEMA.mentions, SKOS.broadMatch, SAPPHO_PROP.expr_references))
        
    # sappho_prop:references_person / sappho_prop:person_referenced_by
    
    if any(g.triples((None, ECRM.P67_refers_to, ECRM.E21_Person))):
        g.add((SAPPHO_PROP.references_person, RDF.type, OWL.ObjectProperty))
        g.add((SAPPHO_PROP.references_person, RDFS.label,
            Literal("Reference to person", lang="en")))
        chain_bnode = BNode()
        Collection(g, chain_bnode, [
            INTRO.R18_showsActualization,
            ECRM.P67_refers_to
        ])
        g.add((SAPPHO_PROP.references_person, OWL.propertyChainAxiom, chain_bnode))
        
        g.add((SAPPHO_PROP.references_person, RDFS.domain, LRMOO.F2_Expression))
        g.add((SAPPHO_PROP.references_person, RDFS.range, ECRM.E21_Person))
        g.add((SAPPHO_PROP.references_person, RDFS.subPropertyOf, ECRM.P67_refers_to))

        g.add((SAPPHO_PROP.person_referenced_by, RDF.type, OWL.ObjectProperty))
        g.add((SAPPHO_PROP.person_referenced_by, RDFS.label,
            Literal("Person referenced by expression", lang="en")))
        g.add((SAPPHO_PROP.person_referenced_by, OWL.inverseOf, SAPPHO_PROP.references_person))
        g.add((SAPPHO_PROP.person_referenced_by, RDFS.domain, LRMOO.F2_Expression))
        g.add((SAPPHO_PROP.person_referenced_by, RDFS.range,  ECRM.E21_Person))
        g.add((SAPPHO_PROP.person_referenced_by, RDFS.subPropertyOf, ECRM.P67i_is_referred_to_by))
            
        g.add((ONTOPOETRY_CORE.mentionsAgent, SKOS.broadMatch, SAPPHO_PROP.references_person))
        g.add((ONTOPOETRY_CORE.isAgentMentionedIn, OWL.inverseOf, ONTOPOETRY_CORE.mentionsAgent))
    
    # sappho_prop:references_place / sappho_prop:place_referenced_by
    
    if any(g.triples((None, ECRM.P67_refers_to, ECRM.E53_Place))):
        g.add((SAPPHO_PROP.references_place, RDF.type, OWL.ObjectProperty))
        g.add((SAPPHO_PROP.references_place, RDFS.label,
            Literal("Reference to place", lang="en")))
        chain_bnode = BNode()
        Collection(g, chain_bnode, [
            INTRO.R18_showsActualization,
            ECRM.P67_refers_to
        ])
        g.add((SAPPHO_PROP.references_place, OWL.propertyChainAxiom, chain_bnode))
        
        g.add((SAPPHO_PROP.references_place, RDFS.domain, LRMOO.F2_Expression))
        g.add((SAPPHO_PROP.references_place, RDFS.range, ECRM.E53_Place))
        g.add((SAPPHO_PROP.references_place, RDFS.subPropertyOf, ECRM.P67_refers_to))
        
        g.add((SAPPHO_PROP.place_referenced_by, RDF.type, OWL.ObjectProperty))
        g.add((SAPPHO_PROP.place_referenced_by, RDFS.label,
            Literal("Place referenced by expression", lang="en")))
        g.add((SAPPHO_PROP.place_referenced_by, OWL.inverseOf, SAPPHO_PROP.references_place))
        g.add((SAPPHO_PROP.place_referenced_by, RDFS.domain, LRMOO.F2_Expression))
        g.add((SAPPHO_PROP.place_referenced_by, RDFS.range,  ECRM.E53_Place))
        g.add((SAPPHO_PROP.place_referenced_by, RDFS.subPropertyOf, ECRM.P67i_is_referred_to_by))
        
        g.add((ONTOPOETRY_CORE.mentionsPlace, SKOS.broadMatch, SAPPHO_PROP.references_place))
        g.add((ONTOPOETRY_CORE.isPlaceMentionedIn, OWL.inverseOf, ONTOPOETRY_CORE.mentionsPlace))

    # sappho_prop:has_character / sappho_prop:is_character_in: link character and expression
    if any(g.triples((None, RDF.type, INTRO.INT_Character))):
        properties = [
            ("has_character",    GOLEM.GP1i_has_character),
            ("is_character_in",  GOLEM.GP1i_is_character_in),
        ]

        for local_name, golem_prop in properties:
            prop = SAPPHO_PROP[local_name]
            g.add((prop, RDF.type, OWL.ObjectProperty))
            g.add((prop, RDFS.label, Literal(local_name, lang="en")))
            g.add((prop, SKOS.closeMatch, golem_prop))

            if local_name == "has_character":
                g.add((prop, RDFS.domain, LRMOO.F2_Expression))
                g.add((prop, RDFS.range,  INTRO.INT2_ActualizationOfFeature))
                g.add((prop, RDFS.subPropertyOf, ECRM.P148_has_component))
            else:
                g.add((prop, RDFS.domain, INTRO.INT2_ActualizationOfFeature))
                g.add((prop, RDFS.range,  LRMOO.F2_Expression))
                g.add((prop, RDFS.subPropertyOf, ECRM.P148i_is_component_of))

        for expr in g.subjects(RDF.type, LRMOO.F2_Expression):
            for act in g.objects(expr, INTRO.R18_showsActualization):
                feat = g.value(act, INTRO.R17_actualizesFeature)
                if feat and (feat, RDF.type, INTRO.INT_Character) in g:
                    g.add((expr, SAPPHO_PROP.has_character,    act))
                    g.add((act,  SAPPHO_PROP.is_character_in, expr))
        
        g.add((ONTOPOETRY_CORE.characterIn, SKOS.closeMatch, SAPPHO_PROP.is_character_in))
        g.add((ONTOPOETRY_CORE.hasCharacter, SKOS.closeMatch, SAPPHO_PROP.has_character))
        g.add((SCHEMA.character, SKOS.closeMatch, SAPPHO_PROP.has_character))
    
    # Serialize
    g.serialize(destination=str(args.output), format="turtle")

    # Remove DBpedia prefixes in-place
    ttl_path = Path(args.output)
    text = ttl_path.read_text(encoding="utf-8")

    text = re.sub(r'^@prefix\s+dbpedia:\s*<[^>]+>\s*\.\s*\n', '', text, flags=re.MULTILINE)

    text = re.sub(r'\bdbpedia:([A-Za-z0-9_/]+)\b', r'<https://dbpedia.org/\1>', text)

    ttl_path.write_text(text, encoding="utf-8")
    print(f"✅ File saved as {args.output}")

if __name__ == "__main__":
    main()



from importlib import resources
from pathlib import Path
from typing import Iterable

def shapes_path(*parts: str) -> Path:
    """Return a pathlib.Path to a file inside the installed shapes directory."""
    base = resources.files(__package__) / "shapes"
    return Path(base) / Path(*parts)



"""
This script retrieves person data from Wikidata based on a list of QIDs (from a CSV file)
and transforms it into CIDOC CRM (OWL/eCRM) RDF triples.

The output is serialized as Turtle and written to 'authors.ttl'.
"""

import csv
import time
import logging
import rdflib
from rdflib import Graph, Literal, Namespace, URIRef
from rdflib.namespace import RDF, RDFS, OWL, XSD
import requests
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry
from email.utils import parsedate_to_datetime
from datetime import datetime, timezone
from typing import Optional, Dict, Any, Iterable, List
from pathlib import Path
from tqdm import tqdm
import argparse
from pyshacl import validate
from wiki2crm import resources

# Settings
SPARQL_ENDPOINT = "https://query.wikidata.org/sparql"
USER_AGENT = "SapphoDataIntegrationBot/1.0 (laura.untner@fu-berlin.de)"
HTTP_TIMEOUT = 90
MAX_RETRIES = 5

# Namespaces
CRM = Namespace("http://www.cidoc-crm.org/cidoc-crm/") # CIDOC CRM
ECRM = Namespace("http://erlangen-crm.org/current/")  # eCRM - CIDOC CRM (OWL version)
ECRM_URI = URIRef("http://erlangen-crm.org/current/")
PROV = Namespace("http://www.w3.org/ns/prov#")  # PROV-O - Provenance Ontology
WD = "http://www.wikidata.org/entity/"  # Base URI for Wikidata entities
SAPPHO_BASE_URI = "https://sappho-digital.com/"  # Base URI for Sappho
SAPPHO = Namespace("https://sappho-digital.com/")

# HTTP helpers
def _parse_retry_after(header_val: str) -> Optional[float]:
    """
    Parse 'Retry-After' per RFC 7231. It may be either a delay in seconds
    or an HTTP-date. Returns seconds as float, or None if not parseable.
    """
    if not header_val:
        return None
    header_val = header_val.strip()
    if header_val.isdigit():
        return float(header_val)
    try:
        dt = parsedate_to_datetime(header_val)
        if dt.tzinfo is None:
            dt = dt.replace(tzinfo=timezone.utc)
        return max(0.0, (dt - datetime.now(timezone.utc)).total_seconds())
    except Exception:
        return None

def make_session() -> requests.Session:
    """
    Build a pooled Session. We keep manual control over retries to fully respect Retry-After.
    """
    sess = requests.Session()
    sess.headers.update({"User-Agent": USER_AGENT})
    retry = Retry(
        total=0,  # manual control below
        respect_retry_after_header=True,
        backoff_factor=0,
        status_forcelist=(429, 500, 502, 503, 504),
        allowed_methods=frozenset(["GET", "POST"]),
        raise_on_status=False,
    )
    adapter = HTTPAdapter(max_retries=retry, pool_connections=10, pool_maxsize=20)
    sess.mount("http://", adapter)
    sess.mount("https://", adapter)
    return sess

SESSION = make_session()

def http_request_with_retry(
    method: str,
    url: str,
    *,
    params: Optional[Dict[str, Any]] = None,
    data: Optional[Dict[str, Any]] = None,
    headers: Optional[Dict[str, str]] = None,
    ok_statuses: Iterable[int] = (200,),
    max_retries: int = MAX_RETRIES,
    timeout: int = HTTP_TIMEOUT,
) -> requests.Response:
    """
    Perform an HTTP request and retry on 429/5xx.
    Uses Retry-After if present; otherwise applies a gentle backoff.
    """
    tries = 0
    while True:
        tries += 1
        resp = SESSION.request(method, url, params=params, data=data, headers=headers, timeout=timeout)

        if resp.status_code in ok_statuses:
            return resp

        if resp.status_code == 429:
            retry_after = _parse_retry_after(resp.headers.get("Retry-After", ""))
            wait_s = retry_after if retry_after is not None else 5.0
            logging.warning("429 Too Many Requests – waiting %.2fs (try %d/%d)", wait_s, tries, max_retries)
            if tries >= max_retries:
                resp.raise_for_status()
            time.sleep(wait_s)
            continue

        if 500 <= resp.status_code < 600:
            retry_after = _parse_retry_after(resp.headers.get("Retry-After", "")) or min(10.0, 1.5 ** (tries - 1))
            logging.warning("%s Server error – waiting %.2fs (try %d/%d)", resp.status_code, retry_after, tries, max_retries)
            if tries >= max_retries:
                resp.raise_for_status()
            time.sleep(retry_after)
            continue

        # non-retriable error
        resp.raise_for_status()

def query_wikidata_sparql(query: str, accept: str = "application/sparql-results+json") -> dict:
    """
    Execute a SPARQL query against Wikidata using the retry-aware HTTP routine.
    """
    headers = {"Accept": accept}
    resp = http_request_with_retry(
        "GET",
        SPARQL_ENDPOINT,
        params={"query": query},
        headers=headers,
    )
    return resp.json()

# Ontology, graph creation, and bindings
def create_graph() -> Graph:
    # Create the RDF graph
    g = Graph()
    g.bind("crm", CRM)
    g.bind("ecrm", ECRM)
    g.bind("prov", PROV)
    g.bind("owl", OWL)
    g.bind("rdfs", RDFS)
    g.bind("sappho", SAPPHO)

    # Ontology
    ontology_uri = URIRef("https://sappho-digital.com/ontology/authors")
    g.add((ontology_uri, RDF.type, OWL.Ontology))
    g.add((ontology_uri, OWL.imports, ECRM_URI))

    # CIDOC CRM alignment and property inverses
    ecrm_to_crm = [
        # Classes
        "E21_Person", "E67_Birth", "E69_Death", "E52_Time-Span", "E53_Place",
        "E36_Visual_Item", "E38_Image", "E55_Type", "E42_Identifier", "E82_Actor_Appellation"
    ]

    for cls in ecrm_to_crm:
        g.add((ECRM.term(cls), OWL.equivalentClass, CRM.term(cls)))

    # Properties
    ecrm_properties = [
        ("P1_is_identified_by", "P1i_identifies"),
        ("P2_has_type", "P2i_is_type_of"),
        ("P4_has_time-span", "P4i_is_time-span_of"),
        ("P7_took_place_at", "P7i_witnessed"),
        ("P65_shows_visual_item", "P65i_is_shown_by"),
        ("P98_brought_into_life", "P98i_was_born"),
        ("P100_was_death_of", "P100i_died_in"),
        ("P131_is_identified_by", "P131i_identifies"),
        ("P138_represents", "P138i_has_representation")
    ]

    for direct, inverse in ecrm_properties:
        g.add((ECRM.term(direct), OWL.inverseOf, ECRM.term(inverse)))
        g.add((ECRM.term(direct), OWL.equivalentProperty, CRM.term(direct)))
        g.add((ECRM.term(inverse), OWL.inverseOf, ECRM.term(direct)))
        g.add((ECRM.term(inverse), OWL.equivalentProperty, CRM.term(inverse)))

    return g

# Function to get Wikidata data in batches
def get_wikidata_batch(qids: List[str], max_retries: int = MAX_RETRIES) -> Dict[str, List[dict]]:
    values = " ".join(f"wd:{qid}" for qid in qids)
    endpoint = SPARQL_ENDPOINT
    query = f"""
    SELECT ?item ?itemLabel ?gender ?genderLabel ?birthPlace ?birthPlaceLabel ?birthDate ?deathPlace ?deathPlaceLabel ?deathDate ?image WHERE {{
      VALUES ?item {{ {values} }}
      OPTIONAL {{ ?item wdt:P21 ?gender . }}
      OPTIONAL {{ ?item wdt:P569 ?birthDate . }}
      OPTIONAL {{ ?item wdt:P19 ?birthPlace . }}
      OPTIONAL {{ ?item wdt:P570 ?deathDate . }}
      OPTIONAL {{ ?item wdt:P20 ?deathPlace . }}
      OPTIONAL {{ ?item wdt:P18 ?image . }}
      SERVICE wikibase:label {{
        bd:serviceParam wikibase:language "en" .
      }}
    }}
    """
    headers = {
        "Accept": "application/sparql-results+json",
        "User-Agent": USER_AGENT
    }

    tries = 0
    while True:
        tries += 1
        try:
            resp = http_request_with_retry(
                "GET",
                endpoint,
                params={"query": query},
                headers=headers,
                ok_statuses=(200,),
                max_retries=max_retries,
                timeout=HTTP_TIMEOUT,
            )
            results = resp.json()["results"]["bindings"]
            grouped: Dict[str, List[dict]] = {}
            for b in results:
                uri = b["item"]["value"]
                qid = uri.split("/")[-1]
                grouped.setdefault(qid, []).append(b)
            return grouped
        except requests.exceptions.RequestException as e:
            # The underlying http_request_with_retry already waited per Retry-After.
            # If we still get here, we apply a small extra linear wait before next outer try.
            wait = min(5.0 * tries, 20.0)
            print(f"[RETRY {tries}] Batch request failed: {e} – retrying in {wait:.1f}s...")
            if tries >= max_retries:
                print("[ERROR] Maximum retries reached. Skipping batch.")
                return {}
            time.sleep(wait)

def load_qids(path: Path) -> List[str]:
    """
    Load QIDs from the given CSV (expects one QID per row, keep original semantics).
    """
    all_qids: List[str] = []
    with path.open(newline="", encoding="utf-8") as csvfile:
        reader = csv.reader(csvfile)
        for row in reader:
            if not row:
                continue
            qid = row[0].strip()
            if qid.startswith("Q"):
                all_qids.append(qid)
    return all_qids

def format_date(iso_string: str) -> str:
    return iso_string.split("T")[0]

def process_authors(g: Graph, all_qids: List[str]) -> None:
    """
    Process QIDs in batches and populate the graph (unchanged triple logic).
    """
    batch_size = 20
    gender_cache: Dict[str, URIRef] = {}
    place_cache: Dict[str, URIRef] = {}  # kept for parity, though not used for memo here
    time_span_cache: Dict[URIRef, URIRef] = {}

    for i in tqdm(range(0, len(all_qids), batch_size)):
        batch = all_qids[i:i+batch_size]
        batch_data = get_wikidata_batch(batch)
        for qid in batch:
            uri = f"http://www.wikidata.org/entity/{qid}"
            bindings = batch_data.get(qid, [])
            if not bindings:
                continue

            b = bindings[0]
            label = b.get("itemLabel", {}).get("value", "").strip()
            if not label:
                label = f"Unknown ({qid})"

            person_uri = URIRef(f"{SAPPHO_BASE_URI}person/{qid}")
            name_uri = URIRef(f"{SAPPHO_BASE_URI}appellation/{qid}")
            identifier_uri = URIRef(f"{SAPPHO_BASE_URI}identifier/{qid}")

            # Person core data
            g.add((person_uri, RDF.type, ECRM.E21_Person))
            g.add((person_uri, OWL.sameAs, URIRef(uri)))
            g.add((person_uri, ECRM.P131_is_identified_by, name_uri))
            g.add((name_uri, ECRM.P131i_identifies, person_uri))
            g.add((name_uri, RDF.type, ECRM.E82_Actor_Appellation))
            g.add((name_uri, RDFS.label, Literal(label, lang="en")))
            g.add((name_uri, PROV.wasDerivedFrom, URIRef(uri)))
            g.add((person_uri, RDFS.label, Literal(label, lang="en")))

            g.add((person_uri, ECRM.P1_is_identified_by, identifier_uri))
            g.add((identifier_uri, ECRM.P1i_identifies, person_uri))
            g.add((identifier_uri, RDF.type, ECRM.E42_Identifier))
            g.add((identifier_uri, RDFS.label, Literal(qid)))
            g.add((identifier_uri, ECRM.P2_has_type, URIRef(f"{SAPPHO_BASE_URI}id_type/wikidata")))
            g.add((URIRef(f"{SAPPHO_BASE_URI}id_type/wikidata"), ECRM.P2i_is_type_of, identifier_uri))
            g.add((URIRef(f"{SAPPHO_BASE_URI}id_type/wikidata"), RDF.type, ECRM.E55_Type))
            g.add((URIRef(f"{SAPPHO_BASE_URI}id_type/wikidata"), RDFS.label, Literal("Wikidata ID", lang="en")))

            def create_timespan_uri(date_value: str) -> URIRef:
                return URIRef(f"{SAPPHO_BASE_URI}timespan/{date_value.replace('-', '')}")

            for event_type, date_key, place_key, class_uri, inverse_prop, direct_prop in [
                ("birth", "birthDate", "birthPlace", ECRM.E67_Birth, ECRM.P98i_was_born, ECRM.P98_brought_into_life),
                ("death", "deathDate", "deathPlace", ECRM.E69_Death, ECRM.P100i_died_in, ECRM.P100_was_death_of)
            ]:
                has_date = date_key in b
                has_place = place_key in b
                if has_date or has_place:
                    event_uri = URIRef(f"{SAPPHO_BASE_URI}{event_type}/{qid}")
                    g.add((person_uri, inverse_prop, event_uri))
                    g.add((event_uri, direct_prop, person_uri))
                    g.add((event_uri, RDF.type, class_uri))
                    g.add((event_uri, RDFS.label, Literal(f"{event_type.capitalize()} of {label}", lang="en")))
                    g.add((event_uri, PROV.wasDerivedFrom, URIRef(uri)))

                    if has_date:
                        date_value = format_date(b[date_key]["value"])
                        date_uri = create_timespan_uri(date_value)
                        if date_uri not in time_span_cache:
                            g.add((date_uri, RDF.type, ECRM.term("E52_Time-Span")))
                            g.add((date_uri, RDFS.label, Literal(date_value, datatype=XSD.date)))
                            time_span_cache[date_uri] = date_uri
                        g.add((event_uri, ECRM["P4_has_time-span"], date_uri))
                        g.add((date_uri, ECRM["P4i_is_time-span_of"], event_uri))

                    if has_place:
                        wikidata_place_uri = b[place_key]["value"]
                        place_id = wikidata_place_uri.split("/")[-1]
                        place_uri = URIRef(f"{SAPPHO_BASE_URI}place/{place_id}")
                        place_label = b.get(f"{place_key}Label", {}).get("value")
                        g.add((event_uri, ECRM.P7_took_place_at, place_uri))
                        g.add((place_uri, ECRM.P7i_witnessed, event_uri))
                        g.add((place_uri, RDF.type, ECRM.E53_Place))
                        g.add((place_uri, OWL.sameAs, URIRef(wikidata_place_uri)))
                        if place_label:
                            g.add((place_uri, RDFS.label, Literal(place_label, lang="en")))

            gender_uri_raw = b.get("gender", {}).get("value")
            gender_label = b.get("genderLabel", {}).get("value")
            if gender_uri_raw and gender_label:
                if gender_uri_raw not in gender_cache:
                    sappho_gender_uri = URIRef(f"{SAPPHO_BASE_URI}gender/{gender_uri_raw.split('/')[-1]}")
                    g.add((sappho_gender_uri, RDF.type, ECRM.E55_Type))
                    g.add((sappho_gender_uri, RDFS.label, Literal(gender_label, lang="en")))
                    g.add((sappho_gender_uri, OWL.sameAs, URIRef(gender_uri_raw)))
                    g.add((sappho_gender_uri, ECRM.P2_has_type, URIRef(f"{SAPPHO_BASE_URI}gender_type/wikidata")))
                    g.add((
                        URIRef(f"{SAPPHO_BASE_URI}gender_type/wikidata"),
                        ECRM.P2i_is_type_of,
                        sappho_gender_uri
                    ))
                    g.add((URIRef(f"{SAPPHO_BASE_URI}gender_type/wikidata"), RDF.type, ECRM.E55_Type))
                    g.add((URIRef(f"{SAPPHO_BASE_URI}gender_type/wikidata"), RDFS.label, Literal("Wikidata Gender", lang="en")))
                    gender_cache[gender_uri_raw] = sappho_gender_uri
                g.add((person_uri, ECRM.P2_has_type, gender_cache[gender_uri_raw]))
                g.add((gender_cache[gender_uri_raw], ECRM.P2i_is_type_of, person_uri))

            image_url = b.get("image", {}).get("value")
            if image_url:
                image_instance_uri = URIRef(f"{SAPPHO_BASE_URI}image/{qid}")
                visual_item_uri = URIRef(f"{SAPPHO_BASE_URI}visual_item/{qid}")
                g.add((visual_item_uri, RDF.type, ECRM.E36_Visual_Item))
                g.add((visual_item_uri, RDFS.label, Literal(f"Visual representation of {label}", lang="en")))
                g.add((visual_item_uri, ECRM.P138_represents, person_uri))
                g.add((person_uri, ECRM.P138i_has_representation, visual_item_uri))
                g.add((image_instance_uri, RDF.type, ECRM.E38_Image))
                g.add((image_instance_uri, ECRM.P65_shows_visual_item, visual_item_uri))
                g.add((visual_item_uri, ECRM.P65i_is_shown_by, image_instance_uri))
                g.add((image_instance_uri, RDFS.seeAlso, URIRef(image_url)))
                g.add((image_instance_uri, PROV.wasDerivedFrom, URIRef(uri)))

def parse_args(argv=None) -> argparse.Namespace:
    p = argparse.ArgumentParser(
        description="Retrieve person data from Wikidata and emit CIDOC CRM TTL"
    )
    p.add_argument("--input",  type=Path, help="CSV with QIDs (e.g. examples/inputs/author-qids.csv)")
    p.add_argument("--output", type=Path, help="Output Turtle (e.g. examples/outputs/authors.ttl)")
    p.add_argument(
        "--shapes",
        type=Path,
        default=resources.shapes_path("author-shapes.ttl"),
        help="Path to SHACL shapes (default: package-installed author-shapes.ttl)",
    )
    p.add_argument("--log-level", default="INFO", help="Logging level (default: INFO)")
    return p.parse_args(argv)

def main(argv=None) -> int:
    args = parse_args(argv)

    logging.basicConfig(level=getattr(logging, args.log_level.upper(), logging.INFO),
                        format="%(levelname)s:%(name)s:%(message)s")

    if args.input is None:
        repo_input = Path("examples/inputs/author-qids.csv")
        if repo_input.exists():
            args.input = repo_input
        else:
            raise SystemExit("--input is required (no examples/inputs/author-qids.csv found)")

    if args.output is None:
        repo_outdir = Path("examples/outputs")
        args.output = (repo_outdir / "authors.ttl") if repo_outdir.exists() else Path("authors.ttl")

    args.output.parent.mkdir(parents=True, exist_ok=True)

    # Create graph
    g = create_graph()

    # Load QIDs
    all_qids = load_qids(args.input)

    # Process and enrich graph
    process_authors(g, all_qids)

    # Serialize to Turtle
    g.serialize(destination=str(args.output), format="turtle")
    print(f"✅ RDF graph written to {args.output}")

    # Validate the output graph using pySHACL
    shapes_graph = Graph().parse(str(args.shapes), format="turtle")

    conforms, report_graph, report_text = validate(
        g,
        shacl_graph=shapes_graph,
        inference="rdfs",
        meta_shacl=True,
        advanced=True,
        abort_on_first=False,
        debug=False,
    )

    print("\n🔎 SHACL Validation Report:")
    print(report_text)

    if not conforms:
        print("❌ Validation failed.")
    else:
        print("✅ Data conforms to SHACL shapes.")

    return 0

if __name__ == "__main__":
    raise SystemExit(main())